"""todo"""

from typing import Dict, List, Set

import pandas as pd

from organon.fl.core.businessobjects.dataframe import DataFrame
from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.extensions.string_extensions import to_upper_eng
from organon.fl.core.helpers.data_frame_helper import get_column_native_type
from organon.fl.mathematics.helpers.pandas_dataframe_operations import PandasDataFrameOperations
from organon.idq.dataaccess.helpers.record_source_filter_helper import RecordSourceFilterHelper
from organon.idq.domain.businessobjects.data_column.dq_file_data_column import DqFileDataColumn
from organon.idq.domain.businessobjects.dq_data_column_collection import DqDataColumnCollection
from organon.idq.domain.businessobjects.dq_data_source import DqDataSource
from organon.idq.domain.services.statistics.source_statistics.dq_source_stats_base_service import \
    DqSourceStatsBaseService
from organon.idq.domain.settings.calculation.dq_file_calculation_parameters import DqFileCalculationParameters
from organon.idq.domain.settings.dq_column_metadata import DqColumnMetadata


class FileSourceStatsService(DqSourceStatsBaseService[DqFileCalculationParameters]):
    """Service for generating statistics for a dq file source"""

    def __init__(self, calc_params: DqFileCalculationParameters):
        super().__init__(calc_params)
        self.record_source = calc_params.input_source_settings.source.locator

    def _get_column_cardinalities_from_population(self, column_names: List[str]) -> Dict[str, int]:
        distinct_values_by_col: Dict[str, Set] = {col: set() for col in column_names}
        for data_frame in RecordSourceFilterHelper.read_csv_with_chunks(self.calc_params.input_source_settings):
            frame = DataFrame()
            frame.data_frame = data_frame
            for col in column_names:
                distinct_values_by_col[col].update(
                    PandasDataFrameOperations.get_unique_values_in_column(frame, col))
        return {col: len(distinct_values) for col, distinct_values in distinct_values_by_col.items()}

    def _get_histograms_for_columns(self, dq_data_columns: Dict[str, DqFileDataColumn]) \
            -> Dict[str, Dict[str, int]]:
        column_names = list(dq_data_columns.keys())
        all_frequencies: Dict[str, Dict[str, int]] = {col: {} for col in column_names}

        for data_frame in RecordSourceFilterHelper.read_csv_with_chunks(self.calc_params.input_source_settings):
            frame = DataFrame()
            frame.data_frame = data_frame
            for col in column_names:
                col_native_type = dq_data_columns[col].column_native_type
                col_frequencies = PandasDataFrameOperations.get_frequencies(frame, col)
                if col_native_type == ColumnNativeType.Numeric:
                    col_frequencies = {str(key): val for key, val in col_frequencies.items()}
                elif col_native_type == ColumnNativeType.Date:
                    new_dict = {}
                    for val, count in col_frequencies.items():
                        date_val: pd.Timestamp = val
                        new_dict[date_val.strftime("%Y%m%d")] = count
                    col_frequencies = new_dict
                current_frequencies_for_column = all_frequencies[col]
                for value, count in col_frequencies.items():
                    current_count = current_frequencies_for_column.get(value, None)
                    if current_count is not None:
                        current_frequencies_for_column[value] += count
                    else:
                        current_frequencies_for_column[value] = count

        return all_frequencies

    def get_data_column_collection(self, data_source: DqDataSource, columns_dq_metadata: List[DqColumnMetadata]) \
            -> DqDataColumnCollection:
        collection = DqDataColumnCollection()
        dq_metadata_dict: Dict[str, DqColumnMetadata] = \
            {to_upper_eng(col.column_name): col for col in columns_dq_metadata}
        for col in data_source.sampled_data.columns:
            dq_data_col = DqFileDataColumn()
            dq_data_col.column_name = col
            dq_data_col.col_np_dtype = data_source.sampled_data.dtypes[col]
            dq_metadata = dq_metadata_dict.get(to_upper_eng(col), None)
            if dq_metadata is not None:
                dq_data_col.default_values = dq_metadata.default_values
                dq_data_col.inclusion_flag = dq_metadata.inclusion_flag
            dq_data_col.column_native_type = get_column_native_type(data_source.sampled_data, col)
            collection.add(dq_data_col)
        return collection



"""
This module includes enum class BinaryTargetClass.
"""
from enum import Enum


# pylint: disable=invalid-name
class BinaryTargetClass(Enum):
    """
    BinaryTargetClass
    """
    NAN = -1
    POSITIVE = 0
    NEGATIVE = 1
    EXCLUSION = 2
    INDETERMINATE = 3



"""todo"""
from organon.idq.domain.businessobjects.dq_data_column_collection import DqDataColumnCollection


class DataSourceStatistics:
    """todo"""

    def __init__(self):
        self.row_count: int = None
        self.size_in_bytes: int = None
        self.data_column_collection: DqDataColumnCollection = None



"""Includes CrossValidationOutput class"""
from typing import List

import numpy as np

from organon.ml.modelling.algorithms.core.abstractions.base_modeller import BaseModeller


class CrossValidationOutput:
    """Output of CrossValidation"""

    def __init__(self):
        self.fold_scores: List[List[float]] = None
        self.mean_scores: List[float] = None
        self.best_modeller: BaseModeller = None
        self.test_fold: np.ndarray = None



"""todo"""
from organon.idq.domain.enums.dq_test_group_type import DqTestGroupType


class DqTestGroup:
    """todo"""

    def __init__(self):
        self.group_type: DqTestGroupType = None
        self.min_bmh: int = None
        self.test_bmh: int = None
        self.inclusion_flag: bool = None



""" This module includes StabilityFeatureReduction"""
from organon.ml.feature_reduction.domain.enums.feature_reduction_types import FeatureReductionType
from organon.ml.feature_reduction.domain.objects.feature_reduction_output import FeatureReductionOutput
from organon.ml.feature_reduction.domain.reductions.base_feature_reduction import BaseFeatureReduction
from organon.ml.feature_reduction.settings.objects.stability_feature_reduction_settings import \
    StabilityFeatureReductionSettings


class StabilityFeatureReduction(BaseFeatureReduction):
    """StabilityFeatureReduction class"""

    def _execute_reduction(self, settings: StabilityFeatureReductionSettings) -> FeatureReductionOutput:
        output = FeatureReductionOutput()
        output.feature_reduction_type = FeatureReductionType.STABILITY

        included_columns = self._get_included_columns(settings.data, settings.excluded_columns)

        if not included_columns:
            output.reduced_column_list = None
            return output

        reduced_columns = [col for col in included_columns if len(settings.data[col].value_counts()) == 1]

        output.reduced_column_list = reduced_columns
        return output

    def get_description(self) -> str:
        return "Stability Feature Reduction"



"""Includes ObjectClassificationServiceFactory class"""
import os
from typing import Union, TYPE_CHECKING

from organon.ml.object_classification.common.helpers import throw_val_ex_with_log, check_directory
from organon.ml.object_classification.domain.common.helpers import get_data_source_settings
from organon.ml.object_classification.domain.data_sources.labeled_dir_data_source_handler import \
    LabeledDirDataSourceHandler
from organon.ml.object_classification.domain.data_sources.tf_dataset_data_source_handler import \
    TfDatasetDataSourceHandler
from organon.ml.object_classification.domain.data_sources.unlabeled_dir_data_source_handler import \
    UnlabeledDirDataSourceHandler
from organon.ml.object_classification.domain.objects.object_clf_settings import ObjectClfSettings

if TYPE_CHECKING:
    import tensorflow as tf  # noqa


class DataSourceHandlerFactory:
    """Class for deciding which service is going to be used in object classification (Binary-Multiclass)"""

    @staticmethod
    def get_data_source_handler(data: Union["tf.data.Dataset", str], settings: ObjectClfSettings) -> Union[
            LabeledDirDataSourceHandler, UnlabeledDirDataSourceHandler, TfDatasetDataSourceHandler]:
        """Returns proper data source handler for given data"""
        settings = get_data_source_settings(settings)
        if isinstance(data, str):
            check_directory(data)
            directory_types = [os.path.isfile(os.path.join(data, directory)) for directory in os.listdir(data)]
            if all(is_file is True for is_file in directory_types):
                data_source = UnlabeledDirDataSourceHandler(data, settings)
            elif all(is_file is False for is_file in directory_types):
                data_source = LabeledDirDataSourceHandler(data, settings)
            else:
                return throw_val_ex_with_log(
                    "Prediction path should contain only unlabeled images (files) or labeled images (folders)")
        else:
            data_source = TfDatasetDataSourceHandler(data, settings)

        return data_source



"""todo"""

from typing import Dict, List

import pandas as pd

from organon.fl.core.businessobjects.dataframe import DataFrame
from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.helpers.data_frame_helper import get_column_native_type
from organon.fl.core.helpers.string_helper import to_upper_eng
from organon.fl.mathematics.helpers.pandas_dataframe_operations import PandasDataFrameOperations
from organon.idq.dataaccess.helpers.record_source_filter_helper import RecordSourceFilterHelper
from organon.idq.domain.businessobjects.data_column.dq_df_data_column import DqDfDataColumn
from organon.idq.domain.businessobjects.dq_data_column_collection import DqDataColumnCollection
from organon.idq.domain.businessobjects.dq_data_source import DqDataSource
from organon.idq.domain.services.statistics.source_statistics.dq_source_stats_base_service import \
    DqSourceStatsBaseService
from organon.idq.domain.settings.calculation.dq_df_calculation_parameters import DqDfCalculationParameters
from organon.idq.domain.settings.dq_column_metadata import DqColumnMetadata


class DfSourceStatsService(DqSourceStatsBaseService[DqDfCalculationParameters]):
    """Service for generating statistics for a dq dataframe source"""

    def __init__(self, calc_params: DqDfCalculationParameters):
        super().__init__(calc_params)
        self.record_source = RecordSourceFilterHelper.get_filtered_dataframe(
            self.calc_params.input_source_settings.source.locator,
            self.calc_params.input_source_settings)

    def _get_column_cardinalities_from_population(self, column_names: List[str]) -> Dict[str, int]:
        frame = DataFrame()
        frame.data_frame = self.record_source
        return PandasDataFrameOperations.get_column_distinct_counts(frame, column_names)

    def _get_histograms_for_columns(self, dq_data_columns: Dict[str, DqDfDataColumn]) \
            -> Dict[str, Dict[str, int]]:
        frame = DataFrame()
        frame.data_frame = self.record_source
        column_names = list(dq_data_columns.keys())
        all_frequencies = {}
        for col in column_names:
            col_native_type = dq_data_columns[col].column_native_type
            col_frequencies = PandasDataFrameOperations.get_frequencies(frame, col)
            if col_native_type == ColumnNativeType.Numeric:
                col_frequencies = {str(key): val for key, val in col_frequencies.items()}
            elif col_native_type == ColumnNativeType.Date:
                new_dict = {}
                for val, count in col_frequencies.items():
                    date_val: pd.Timestamp = val
                    new_dict[date_val.strftime("%Y%m%d")] = count
                col_frequencies = new_dict
            all_frequencies[col] = col_frequencies
        return all_frequencies

    def get_data_column_collection(self, data_source: DqDataSource, columns_dq_metadata: List[DqColumnMetadata]) \
            -> DqDataColumnCollection:
        collection = DqDataColumnCollection()
        dq_metadata_dict: Dict[str, DqColumnMetadata] = \
            {to_upper_eng(col.column_name): col for col in columns_dq_metadata}
        for col in data_source.sampled_data.columns:
            dq_data_col = DqDfDataColumn()
            dq_data_col.column_name = col
            dq_data_col.col_np_dtype = data_source.sampled_data.dtypes[col]
            dq_metadata = dq_metadata_dict.get(to_upper_eng(col), None)
            if dq_metadata is not None:
                dq_data_col.default_values = dq_metadata.default_values
                dq_data_col.inclusion_flag = dq_metadata.inclusion_flag
            dq_data_col.column_native_type = get_column_native_type(data_source.sampled_data, col)
            collection.add(dq_data_col)
        return collection



"""Includes TransferLearningSettings class"""
from organon.ml.common.enums.optimizer_type import OptimizerType


class TransferLearningSettings:
    """Settings for transfer learning in object classification"""

    def __init__(self, epoch: int = 20, early_stopping: int = 2, optimizer: OptimizerType = OptimizerType.ADAM,
                 dropout: float = 0.2, learning_rate: float = 0.001):
        self.epoch = epoch
        self.early_stopping = early_stopping
        self.optimizer = optimizer
        self.dropout = dropout
        self.learning_rate = learning_rate



"""Includes MultiLayerPerceptronClassifier class."""
from sklearn.neural_network import MLPClassifier

from organon.ml.modelling.algorithms.core.abstractions.base_sklearn_classifier import BaseSklearnClassifier


class MultiLayerPerceptronClassifier(BaseSklearnClassifier[MLPClassifier]):
    """Wrapper for sklearn MLPClassifier"""

    @property
    def classifier_class(self):
        return MLPClassifier



"""
This module includes BaseAfeSettings class.
"""
from organon.afe.domain.settings.afe_process_settings import AfeProcessSettings


class BaseAfeSettings:
    """Base class for afe modelling/scoring settings"""

    def __init__(self):
        self.process_settings: AfeProcessSettings = None



"""Includes CFSSelectionService class."""
from typing import Dict

import numpy as np
import pandas as pd
from scipy.stats import pointbiserialr

from organon.fl.core.helpers.data_frame_helper import get_correlation_matrix_memory_efficient
from organon.fl.mathematics.constants import DOUBLE_MIN
from organon.ml.common.enums.target_type import TargetType
from organon.ml.feature_selection.domain.objects.cfs_selection_output import CFSSelectionOutput
from organon.ml.feature_selection.domain.objects.priority_queue import PriorityQueue
from organon.ml.feature_selection.domain.objects.settings.cfs_selection_settings import CFSSelectionSettings
from organon.ml.feature_selection.domain.services.base_feature_selection_service import SelectionRunSettings
from organon.ml.feature_selection.domain.services.base_supervised_feature_selection_service import \
    BaseSupervisedFeatureSelectionService


class CFSSelectionService(BaseSupervisedFeatureSelectionService[CFSSelectionSettings, CFSSelectionOutput]):
    """Service class for CFS selection"""

    @classmethod
    def _validate_settings(cls, settings: CFSSelectionSettings):
        if settings.target_type not in [TargetType.BINARY, TargetType.SCALAR]:
            raise ValueError("Correlation based feature selection works only for BINARY or SCALAR targets")

    @classmethod
    def _run(cls, run_settings: SelectionRunSettings) -> CFSSelectionOutput:
        target = cls._get_target_data_as_series_of_numbers(run_settings.settings)
        settings: CFSSelectionSettings = run_settings.settings
        data = run_settings.data
        features = data.columns
        queue = PriorityQueue()

        visited = []
        # counter for backtracks
        n_backtrack = 0
        # limit of backtracks
        max_backtracks = run_settings.settings.max_backtracks
        corr_dict, best_value, best_feature = cls._get_feature_class_correlations(data, target)
        queue.push({best_feature}, best_value)
        best_subset = set()
        corr_matrix = cls._get_correlation_matrix(data, settings.prioritize_performance)
        while not queue.is_empty():
            # get element of queue with highest merit
            subset, priority = queue.pop()

            # check whether the priority of this subset
            # is higher than the current best subset
            if priority < best_value:
                n_backtrack += 1
            else:
                best_value = priority
                best_subset = subset

            # goal condition
            if n_backtrack == max_backtracks:
                break

            # iterate through all features and look of one can
            # increase the merit
            for feature in features:
                if feature in subset:
                    continue
                temp_subset = subset.copy()
                temp_subset.add(feature)

                # check if this subset has already been evaluated
                for node in visited:
                    if node == temp_subset:
                        break
                # if not, ...
                else:
                    # ... mark it as visited
                    visited.append(temp_subset)
                    # ... compute merit
                    merit = cls._get_merit(temp_subset, corr_dict, corr_matrix)
                    # and push it to the queue
                    queue.push(temp_subset, merit)
        output = CFSSelectionOutput()
        output.selected_features = list(best_subset)
        return output

    @classmethod
    def _get_merit(cls, subset: set, corr_dict: Dict[str, float], corr_matrix):
        k = len(subset)
        # average feature-class correlation
        rcf = np.mean([corr_dict[feature] for feature in subset])

        # average feature-feature correlation
        corr = abs(corr_matrix.loc[list(subset), list(subset)])
        rff = corr.unstack().mean()

        return (k * rcf) / np.sqrt(k + k * (k - 1) * rff)

    @classmethod
    def _get_feature_class_correlations(cls, data: pd.DataFrame, target: pd.Series):
        corr_dict = {}
        best_value = DOUBLE_MIN
        best_feature = None
        for feature in data.columns:
            coeff = pointbiserialr(target, data[feature])
            val = abs(coeff.correlation)
            corr_dict[feature] = val
            if val > best_value:
                best_value = val
                best_feature = feature
        return corr_dict, best_value, best_feature

    @classmethod
    def _get_correlation_matrix(cls, data: pd.DataFrame, prioritize_performance: bool):
        if prioritize_performance:
            corr_df = data.corr()
            corr_df.values[np.tril_indices_from(corr_df.values)] = np.nan
            return corr_df
        return get_correlation_matrix_memory_efficient(data, only_half_filled=True, diagonal_values=np.nan)



"""Includes TextClassificationService class for Text Classification"""
import itertools
import os
from typing import Union, Tuple, Optional

import numpy as np
import pandas as pd
import tensorflow as tf
from datasets import Dataset, load_dataset
from sklearn.model_selection import train_test_split

from organon.fl.logging.helpers.log_helper import LogHelper
from organon.ml.text_classification.domain.enums.model_run_type import ModelRunType
from organon.ml.text_classification.domain.enums.classification_languages import ClassificationLanguages
from organon.ml.text_classification.domain.enums.selection_metrics import SelectionMetrics
from organon.ml.text_classification.domain.objects.text_classification_settings import TextClassificationSettings
from organon.ml.text_classification.domain.enums.model_checkpoints import ModelCheckpoints
from organon.ml.text_classification.domain.objects.tokenizer_settings import TokenizerSettings

# pylint: disable=import-outside-toplevel


class TextClassificationService:
    """Text Classification Domain Service"""

    BINARY_CHECKPOINTS = [ModelCheckpoints.BERT_BASE_TR_128, ModelCheckpoints.BERT_BASE_TR_SENTIMENT]
    TURKISH_CHECKPOINTS = [ModelCheckpoints.BERT_BASE_MLINGUAL, ModelCheckpoints.BERT_BASE_TR_128,
                           ModelCheckpoints.BERT_BASE_TR_SENTIMENT]
    ENGLISH_CHECKPOINTS = [ModelCheckpoints.BERT_BASE_MLINGUAL, ModelCheckpoints.ROBERTA_BASE,
                           ModelCheckpoints.BERT_BASE, ModelCheckpoints.DISTILBERT_BASE]
    val_loss_key = 'Loss'
    accuracy_key = 'Accuracy'
    roc_auc_score_key = 'Roc-Auc Score'
    f1_score_key = 'F1-Score'

    def __init__(self, settings: TextClassificationSettings):
        from transformers import logging as transformers_log
        tf.get_logger().setLevel('ERROR')
        transformers_log.set_verbosity_error()
        if tf.config.list_physical_devices('GPU'):
            LogHelper.warning("GPU is available")
        else:
            LogHelper.warning("GPU is NOT available")

        self._settings = settings
        self.is_trained = False
        self.model = None
        self.grid_search_params_dict: dict = None
        self.best_params_dict: dict = None
        self.metrics: dict = None
        self._num_labels: int = None
        self._tokenizer = None
        self._predictions: np.ndarray = None
        self._label_column = "label"

    def predict(self, test_data: Union[str, Tuple[str, Optional[str]], pd.DataFrame]):
        """predict sample from trained model"""
        x_test = self._load_test_data(test_data)
        self._check_trained()
        predictions = self._calculate_preds(x_test)
        class_predictions = np.argmax(predictions, axis=1)
        return class_predictions

    def predict_proba(self, test_data: Union[str, Tuple[str, Optional[str]], pd.DataFrame]):
        """predict probability values for test texts"""
        from sklearn.metrics import roc_auc_score
        from scipy.special import softmax

        x_test = self._load_test_data(test_data)
        self._check_trained()
        predictions = self._calculate_preds(x_test)
        predictions_softmax = softmax(predictions, axis=1)
        if self._num_labels > 2:
            roc_scr = roc_auc_score(x_test[self._label_column], predictions_softmax, multi_class='ovr')
        else:
            roc_scr = roc_auc_score(x_test[self._label_column], predictions_softmax[:, 1],
                                    multi_class='ovr')

        return predictions_softmax, roc_scr

    def fit(self, data: Union[str, Tuple[str, Optional[str]], pd.DataFrame],
            validation_data: Union[str, pd.DataFrame] = None, validation_data_ratio: float = 0.2,
            random_seed: int = None, sample_weights_flag: bool = True):
        """fit the Text Classification Model"""
        train_data, val_data = self._load_data(data, validation_data, validation_data_ratio, random_seed)
        self._num_labels = len(np.unique(train_data[self._label_column]))

        if self._settings.mdl_run_type == ModelRunType.EFFICIENT:
            self._check_language_checkpoint(self._settings.checkpoint)
            TextClassificationService._check_cls_type(self._settings.checkpoint, self._num_labels)
            model_checkpoint = TextClassificationService._get_checkpoint(self._settings.checkpoint)

            tokenized_x_train, tf_train_set, tf_validation_set, train_class_weights = \
                self._training_preprocess(train_data, val_data, model_checkpoint,
                                          self._settings.mdl_param_settings.batch_size)

            optimizer, callback = \
                TextClassificationService._load_optimizer(tokenized_x_train,
                                                          self._settings.mdl_param_settings.batch_size,
                                                          self._settings.mdl_param_settings.epochs,
                                                          self._settings.opt_settings.learning_rate,
                                                          self._settings.opt_settings.early_stopping,
                                                          self._settings.opt_settings.early_stopping_min_delta)

            model, metrics = self._model_run(val_data=val_data, model_checkpoint=model_checkpoint,
                                             tf_train_set=tf_train_set, tf_validation_set=tf_validation_set,
                                             num_epochs=self._settings.mdl_param_settings.epochs,
                                             optimizer=optimizer, callback=callback,
                                             steps_per_epoch=self._settings.opt_settings.steps_per_epoch,
                                             train_class_weights=train_class_weights,
                                             sample_weights_flag=sample_weights_flag)

            self.model = model
            self.is_trained = True
            self.metrics = metrics

        elif self._settings.mdl_run_type == ModelRunType.HIGH_PERFORMANCE:

            if self._settings.grid_src_settings is not None:
                for checkpoint in self._settings.grid_src_settings.models:
                    self._check_language_checkpoint(checkpoint)

                best_model, best_metrics, best_params, params_dict = self._grid_search(train_data, val_data,
                                                                                       sample_weights_flag)
                self.model = best_model
                self.is_trained = True
                self.metrics = best_metrics
                self.grid_search_params_dict = params_dict
                self.best_params_dict = best_params
            else:
                raise ValueError("Grid Search Settings cannot be None.")
        else:
            raise ValueError("Model Run Type has not specified.")
    @staticmethod
    def load_csv_data(train_data_file: str, validation_data_file: str = None, validation_data_ratio: float = 0.2,
                      random_seed: int = None):
        """load text classification data and turn it to dataframe"""
        if os.path.exists(train_data_file):
            train_data = pd.read_csv(train_data_file)
            if os.path.exists(validation_data_file):
                validation_data = pd.read_csv(validation_data_file)
                return train_data, validation_data
            x_train, x_val = train_test_split(train_data, test_size=validation_data_ratio,
                                              random_state=random_seed)
            return x_train, x_val
        raise FileNotFoundError("csv path can not be located.")

    @staticmethod
    def load_dataframe(train_data: pd.DataFrame, validation_data: pd.DataFrame = None,
                       validation_data_ratio: float = 0.2, random_seed: int = None):
        """load text classification data as dataframe"""
        if validation_data is not None:
            if train_data.empty and validation_data.empty:
                raise ValueError("DataFrames can not be empty")
            return train_data, validation_data
        x_train, x_val = train_test_split(train_data, test_size=validation_data_ratio,
                                          random_state=random_seed)
        return x_train, x_val

    @staticmethod
    def load_hf_data(dataset_name: str = None, subset: str = None, validation_data_ratio: float = 0.2,
                     random_seed: int = None):
        """using load_dataset method, create Hugging Face dataframes"""
        if dataset_name is None:
            raise ValueError("dataset name can not be None")
        if subset is not None:
            hf_dataset = load_dataset(dataset_name, subset)
            if 'validation' in hf_dataset.keys():
                train_data = pd.DataFrame(hf_dataset['train'])
                val_data = pd.DataFrame(hf_dataset['validation'])
            else:
                train_data, val_data = train_test_split(pd.DataFrame(hf_dataset['train']),
                                                        test_size=validation_data_ratio,
                                                        random_state=random_seed)
        else:
            hf_dataset = load_dataset(dataset_name)
            if 'validation' in hf_dataset.keys():
                train_data = pd.DataFrame(hf_dataset['train'])
                val_data = pd.DataFrame(hf_dataset['validation'])
            else:
                train_data, val_data = train_test_split(pd.DataFrame(hf_dataset['train']),
                                                        test_size=validation_data_ratio,
                                                        random_state=random_seed)
        del hf_dataset
        return train_data, val_data

    def _load_data(self, data: Union[str, Tuple[str, Optional[str]], pd.DataFrame],
                   validation_data: Union[str, pd.DataFrame] = None, validation_data_ratio: float = 0.2,
                   random_seed: int = None):
        """load and check data properties"""
        inplace_flag = False
        if isinstance(data, str):
            if self._settings.text_column is None:
                raise ValueError("CSV text column name must be specified")
            if self._settings.target_column is None:
                raise ValueError("CSV target column name must be specified")
            train_data, val_data = TextClassificationService.load_csv_data(data, validation_data, validation_data_ratio,
                                                                           random_seed)
            inplace_flag = True
        elif isinstance(data, Tuple):
            if self._settings.text_column is None:
                raise ValueError("HF Data text column name must be specified")
            if self._settings.target_column is None:
                raise ValueError("HF Data target column name must be specified")
            train_data, val_data = TextClassificationService.load_hf_data(data[0], data[1], validation_data_ratio,
                                                                          random_seed)
            inplace_flag = True
        elif isinstance(data, pd.DataFrame):
            if self._settings.text_column is None:
                raise ValueError("DataFrame text column name must be specified")
            if self._settings.target_column is None:
                raise ValueError("DataFrame target column name must be specified")
            train_data, val_data = TextClassificationService.load_dataframe(data, validation_data,
                                                                            validation_data_ratio, random_seed)
        else:
            raise TypeError("Either pass CSV file path, pandas Dataframe or HF dataset keyword to train the model.")
        train_data = TextClassificationService._check_target_column_name(train_data, self._settings.target_column,
                                                                         inplace_flag)
        val_data = TextClassificationService._check_target_column_name(val_data, self._settings.target_column,
                                                                       inplace_flag)

        return train_data, val_data

    def _load_test_data(self, test_data: Union[str, Tuple[str, Optional[str]], pd.DataFrame]):
        """loads test data for predict and predict_proba methods"""
        inplace_flag = False
        if isinstance(test_data, str):
            test_data = pd.read_csv(test_data)
            inplace_flag = True
        elif isinstance(test_data, Tuple):
            test_data = pd.DataFrame(load_dataset(test_data[0], test_data[1], split="test"))
            inplace_flag = True
        elif isinstance(test_data, pd.DataFrame):
            pass
        else:
            raise TypeError("Either pass CSV file path, pandas Dataframe or HF dataset keyword to make predictions.")
        test_data = TextClassificationService._check_target_column_name(test_data, self._settings.target_column,
                                                                        inplace_flag)
        return test_data

    def _check_trained(self):
        """check if model trained, this method will constrain the usage of predict and predict_proba before training"""
        if not self.is_trained:
            raise Exception("Trainer is not trained yet.")

    def _calculate_preds(self, test_data: pd.DataFrame):
        """calculate prediction probabilities with tokenized test set"""
        tokenized_test_set = TextClassificationService._tokenize_data_words \
            (test_data, self._tokenizer,
             self._settings.text_column,
             self._settings.tokenizer_settings,
             False,
             self._settings.mdl_param_settings.batch_size)
        predictions = self.model.predict(tokenized_test_set)['logits']
        return predictions

    def _grid_search(self, train_data: pd.DataFrame, val_data: pd.DataFrame, sample_weights_flag: bool):
        """grid search algorithm for high performance Text Classification"""
        best_model = None
        best_params = []
        grid_search_lists = [
            self._settings.grid_src_settings.models, self._settings.grid_src_settings.batch_sizes,
            self._settings.grid_src_settings.epochs, self._settings.grid_src_settings.learning_rates,
            self._settings.grid_src_settings.early_stopping_patiences,
            self._settings.grid_src_settings.early_stopping_min_deltas
        ]
        best_score = float('-inf')
        best_loss = float('inf')
        params_dict = {}
        best_metrics = {}
        step_id = 1
        for run_step in itertools.product(*grid_search_lists):
            LogHelper.warning(f"Grid Search is running for {run_step[0].name}")
            params_info = "Grid Search Parameters: \n" + f"Checkpoint: {run_step[0].name}\n" + \
                           f"Batch Size: {run_step[1]}\n" + f"Epoch: {run_step[2]}\n" +\
                           f"Learning Rate: {run_step[3]}\n" +\
                           f"Early Stopping Patience: {run_step[4]}\n" +\
                           f"Early Stopping Min Delta: {run_step[5]}\n"
            LogHelper.warning(params_info)
            model_checkpoint = TextClassificationService._get_checkpoint(run_step[0])
            TextClassificationService._check_cls_type(run_step[0], self._num_labels)
            tokenized_x_train, tf_train_set, tf_validation_set, train_class_weights = self._training_preprocess(
                train_data, val_data, model_checkpoint, run_step[1])
            optimizer, callback = TextClassificationService._load_optimizer(tokenized_x_train, batch_size=run_step[1],
                                                                            num_epochs=run_step[2],
                                                                            learning_rate=run_step[3],
                                                                            early_stopping_patience=run_step[4],
                                                                            early_stopping_min_delta=run_step[5]
                                                                            )
            model, selection_metrics = self._model_run(val_data=val_data, model_checkpoint=model_checkpoint,
                                                       tf_train_set=tf_train_set, tf_validation_set=tf_validation_set,
                                                       num_epochs=run_step[2], optimizer=optimizer, callback=callback,
                                                       steps_per_epoch=self._settings.opt_settings.steps_per_epoch,
                                                       train_class_weights=train_class_weights,
                                                       sample_weights_flag=sample_weights_flag)
            LogHelper.warning(
                f"{TextClassificationService.accuracy_key}: {selection_metrics[TextClassificationService.accuracy_key]}"
               f"{TextClassificationService.val_loss_key}: {selection_metrics[TextClassificationService.val_loss_key]}")
            params = {'Checkpoint': run_step[0].name, 'Batch Size': run_step[1],
                      'Epoch': run_step[2], 'Learning Rate': run_step[3],
                      'Early Stopping Patience': run_step[4],
                      'Early Stopping Min Delta': run_step[5],
                      TextClassificationService.accuracy_key: selection_metrics[TextClassificationService.accuracy_key],
                      TextClassificationService.val_loss_key: selection_metrics[TextClassificationService.val_loss_key],
                      TextClassificationService.roc_auc_score_key: selection_metrics[
                          TextClassificationService.roc_auc_score_key],
                      TextClassificationService.f1_score_key: selection_metrics[TextClassificationService.f1_score_key]}
            params_dict[f"Step {step_id}"] = params
            best_metrics, best_model, best_params, best_loss, best_score = self._determine_best_model(model, run_step,
                                                                                                      best_metrics,
                                                                                                      best_model,
                                                                                                      best_params,
                                                                                                      best_loss,
                                                                                                      best_score,
                                                                                                      selection_metrics)
            step_id += 1
        best_params_dict = {'Checkpoint': best_params[0].name, 'Batch Size': best_params[1],
                            'Epoch': best_params[2], 'Learning Rate': best_params[3],
                            'Early Stopping Patience': best_params[4],
                            'Early Stopping Min Delta': best_params[5]}
        return best_model, best_metrics, best_params_dict, params_dict

    def _determine_best_model(self, model, run_step, best_metrics, best_model, best_params, best_loss: float,
                              best_score: float, selection_metrics):
        """determination logic for given metrics"""
        # pylint: disable=too-many-arguments
        if self._settings.grid_src_settings.model_selection_metric == SelectionMetrics.VAL_LOSS:
            if best_loss > selection_metrics[TextClassificationService.val_loss_key]:
                best_loss = selection_metrics[TextClassificationService.val_loss_key]
                best_score = selection_metrics[TextClassificationService.accuracy_key]
                best_model = model
                best_params = run_step
                best_metrics = {TextClassificationService.val_loss_key: best_loss,
                                TextClassificationService.accuracy_key: best_score}
        elif self._settings.grid_src_settings.model_selection_metric == SelectionMetrics.ACC_SCORE:
            if best_score < selection_metrics[TextClassificationService.accuracy_key]:
                best_loss = selection_metrics[TextClassificationService.val_loss_key]
                best_score = selection_metrics[TextClassificationService.accuracy_key]
                best_model = model
                best_params = run_step
                best_metrics = {TextClassificationService.val_loss_key: best_loss,
                                TextClassificationService.accuracy_key: best_score}
        elif self._settings.grid_src_settings.model_selection_metric == SelectionMetrics.ROC_AUC:
            if best_score < selection_metrics[TextClassificationService.roc_auc_score_key]:
                best_loss = selection_metrics[TextClassificationService.val_loss_key]
                best_score = selection_metrics[TextClassificationService.roc_auc_score_key]
                best_model = model
                best_params = run_step
                best_metrics = {TextClassificationService.val_loss_key: best_loss,
                                TextClassificationService.roc_auc_score_key: best_score}
        elif self._settings.grid_src_settings.model_selection_metric == SelectionMetrics.F1_SCORE:
            if best_score < selection_metrics[TextClassificationService.f1_score_key]:
                best_loss = selection_metrics[TextClassificationService.val_loss_key]
                best_score = selection_metrics[TextClassificationService.f1_score_key]
                best_model = model
                best_params = run_step
                best_metrics = {TextClassificationService.val_loss_key: best_loss,
                                TextClassificationService.f1_score_key: best_score}
        else:
            LogHelper.error("Can not determine the best model with given metric.")
        return best_metrics, best_model, best_params, best_loss, best_score

    def _model_run(self, val_data, model_checkpoint: str, tf_train_set, tf_validation_set,
                   num_epochs: int, optimizer, callback, steps_per_epoch: int, train_class_weights,
                   sample_weights_flag: bool):
        # pylint: disable=too-many-arguments
        """runs the specified model"""
        from transformers import TFAutoModelForSequenceClassification
        from scipy.special import softmax

        model = TFAutoModelForSequenceClassification.from_pretrained(model_checkpoint,
                                                                     num_labels=self._num_labels, from_pt=True)

        def add_sample_weights(data):
            # The weights for each class, with the constraint that:
            #     sum(class_weights) == 1.0
            class_weights = tf.constant(list(train_class_weights.values()))
            class_weights = class_weights / tf.reduce_sum(class_weights)

            # Create an image of `sample_weights` by using the label at each pixel as an
            # index into the `class weights` .
            sample_weights = tf.gather(class_weights, indices=tf.cast(data['labels'], tf.int32), axis=0)

            return data, sample_weights

        model.compile(optimizer=optimizer)

        if sample_weights_flag:
            model.fit(x=tf_train_set.map(add_sample_weights),
                      validation_data=tf_validation_set,
                      epochs=num_epochs, callbacks=[callback], steps_per_epoch=steps_per_epoch)
        else:
            model.fit(x=tf_train_set,
                      validation_data=tf_validation_set,
                      epochs=num_epochs, callbacks=[callback], steps_per_epoch=steps_per_epoch)

        loss = model.evaluate(tf_validation_set)
        predictions = model.predict(tf_validation_set)['logits']
        val_predictions_softmax = softmax(predictions, axis=1)
        val_class_predictions = np.argmax(predictions, axis=1)
        val_roc, val_acc, val_f1 = self._calculate_metrics_with_val_data(val_data, val_predictions_softmax,
                                                                         val_class_predictions)
        selection_metrics = {TextClassificationService.val_loss_key: loss,
                             TextClassificationService.roc_auc_score_key: val_roc,
                             TextClassificationService.accuracy_key: val_acc,
                             TextClassificationService.f1_score_key: val_f1}

        return model, selection_metrics

    def _calculate_metrics_with_val_data(self, val_data: pd.DataFrame, predictions_softmax, val_class_predictions):
        """calculate metrics for model evaluation"""
        from sklearn.metrics import roc_auc_score, accuracy_score, f1_score
        if self._num_labels > 2:
            val_roc_auc = roc_auc_score(val_data[self._label_column], predictions_softmax, average="weighted",
                                        multi_class='ovr')
            val_acc = accuracy_score(val_data[self._label_column], val_class_predictions)
            val_f1 = f1_score(val_data[self._label_column], val_class_predictions, average='weighted')
        else:
            val_roc_auc = roc_auc_score(val_data[self._label_column], predictions_softmax[:, 1])
            val_acc = accuracy_score(val_data[self._label_column], val_class_predictions)
            val_f1 = f1_score(val_data[self._label_column], val_class_predictions)

        return val_roc_auc, val_acc, val_f1

    def _training_preprocess(self, train_data: pd.DataFrame, val_data: pd.DataFrame, model_checkpoint: str,
                             batch_size: int):
        """tokenize data for model tranining"""
        from transformers import AutoTokenizer
        from sklearn.utils import class_weight
        self._tokenizer = AutoTokenizer.from_pretrained(model_checkpoint)

        tf_train_set, tokenized_x_train = \
            TextClassificationService._tokenize_data_words(train_data, self._tokenizer,
                                                           self._settings.text_column,
                                                           self._settings.tokenizer_settings,
                                                           shuffle=True,
                                                           batch_size=batch_size)

        tf_validation_set = TextClassificationService._tokenize_data_words(val_data, self._tokenizer,
                                                                           self._settings.text_column,
                                                                           self._settings.tokenizer_settings,
                                                                           shuffle=False, batch_size=batch_size)

        class_weights = class_weight.compute_class_weight(class_weight='balanced',
                                                          classes=np.unique(train_data.label),
                                                          y=train_data.label)

        train_class_weights = dict(enumerate(class_weights))
        return tokenized_x_train, tf_train_set, tf_validation_set, train_class_weights

    @staticmethod
    def _load_optimizer(tokenized_x_train, batch_size: int, num_epochs: int, learning_rate: float,
                        early_stopping_patience: int, early_stopping_min_delta: float):
        """using create optimizer method, make the optimizer ready for training"""
        from keras.callbacks import EarlyStopping
        optimizer = TextClassificationService._create_optimizer(tokenized_x_train, batch_size, num_epochs,
                                                                learning_rate)

        callback = [
            EarlyStopping(patience=early_stopping_patience,
                          min_delta=early_stopping_min_delta, monitor='val_loss')
        ]
        return optimizer, callback

    @staticmethod
    def _tokenize_data_words(data: pd.DataFrame, tokenizer, text_column: str,
                             tokenizer_settings: TokenizerSettings, shuffle: bool, batch_size: int):
        """general purpose dataset tokenization method"""
        from transformers import DataCollatorWithPadding

        def tokenize_function(text_data: pd.DataFrame):
            """tokenize all the text data and return it with the help of the map method"""
            return tokenizer(text_data[text_column], padding=tokenizer_settings.padding,
                             truncation=tokenizer_settings.truncation, max_length=tokenizer_settings.max_length)

        data_d = Dataset.from_pandas(data)
        tokenized_data = data_d.map(tokenize_function, batched=True)
        del data_d
        data_collator = DataCollatorWithPadding(tokenizer=tokenizer, return_tensors='tf')
        data_set = tokenized_data.to_tf_dataset(
            columns=["attention_mask", "input_ids", "label"],
            shuffle=shuffle,  # True for train data, False for both validation and test data
            batch_size=batch_size,
            collate_fn=data_collator
        )
        if shuffle:
            return data_set, tokenized_data
        return data_set

    @staticmethod
    def _get_checkpoint(checkpoint_enum: ModelCheckpoints):
        if checkpoint_enum == ModelCheckpoints.BERT_BASE:
            return "bert-base-cased"
        if checkpoint_enum == ModelCheckpoints.DISTILBERT_BASE:
            return "distilbert-base-uncased"
        if checkpoint_enum == ModelCheckpoints.ROBERTA_BASE:
            return "roberta-base"
        if checkpoint_enum == ModelCheckpoints.BERT_BASE_TR_128:
            return "dbmdz/bert-base-turkish-128k-cased"
        if checkpoint_enum == ModelCheckpoints.BERT_BASE_MLINGUAL:
            return "bert-base-multilingual-cased"
        if checkpoint_enum == ModelCheckpoints.BERT_BASE_TR_SENTIMENT:
            return "savasy/bert-base-turkish-sentiment-cased"

        raise ValueError("Specified checkpoint can not be supported.")

    @staticmethod
    def _create_optimizer(tokenized_x_train, batch_size: int, epochs: int, learning_rate: float):
        """method for model optimizer"""
        from transformers import create_optimizer
        batches_per_epoch = len(tokenized_x_train) // batch_size
        total_train_steps = int(batches_per_epoch * epochs)
        # pylint: disable=assignment-from-no-return
        optimizer, _ = create_optimizer(init_lr=learning_rate, num_warmup_steps=0,
                                        num_train_steps=total_train_steps)

        return optimizer

    @staticmethod
    def _check_cls_type(checkpoint: ModelCheckpoints, num_labels: int):
        """Check if inappropriate checkpoint usage occurs"""
        if num_labels > 2:
            if checkpoint in TextClassificationService.BINARY_CHECKPOINTS:
                raise TypeError("This type of checkpoint can't be used in Multiclass Classification")

    def _check_language_checkpoint(self, checkpoint: ModelCheckpoints):
        """check if selected checkpoint is appropriate for data language"""
        if self._settings.language == ClassificationLanguages.TURKISH:
            if checkpoint not in TextClassificationService.TURKISH_CHECKPOINTS:
                raise ValueError(f"Checkpoint named {checkpoint} is non-Turkish")
        elif self._settings.language == ClassificationLanguages.ENGLISH:
            if checkpoint not in TextClassificationService.ENGLISH_CHECKPOINTS:
                raise ValueError(f"Checkpoint named {checkpoint} is non-English")
        else:
            raise ValueError("Given language cannot be supported.")

    @staticmethod
    def _check_target_column_name(data: pd.DataFrame, target_column: str, inplace_flag: bool):
        """check if target column name set to 'label' if not, set it as 'label'"""
        if target_column != "label":
            if inplace_flag:
                data.rename(columns={target_column: 'label'}, inplace=True)
            else:
                data = data.rename(columns={target_column: 'label'})
        return data



"""
This module includes AfeSettingsReader class.
"""
from organon.afe.domain.common.reader_helper import setattr_if_none
from organon.afe.domain.settings.afe_modelling_settings import AfeModellingSettings
from organon.afe.domain.settings.afe_reading_settings import AfeDataReadingSettings
from organon.afe.domain.settings.base_afe_settings_reader import BaseAfeSettingsReader


class AfeSettingsReader(BaseAfeSettingsReader[AfeModellingSettings]):
    """
    Class for static methods on reading AfeModellingSettings from a file
    """

    def _set_trx_descriptor_defaults(self, modelling_settings: AfeModellingSettings):

        feature_gen_settings = modelling_settings.data_source_settings.trx_descriptor.feature_gen_setting

        if feature_gen_settings.included_operators is None:
            feature_gen_settings.included_operators = self._get_default_included_afe_operators()
        if feature_gen_settings.date_offset is None:
            feature_gen_settings.date_offset = 0
        try:
            trx_descriptor = modelling_settings.data_source_settings.trx_descriptor
            if trx_descriptor.reading_settings is None:
                trx_descriptor.reading_settings = AfeDataReadingSettings()

            setattr_if_none(trx_descriptor.reading_settings, "number_of_rows_per_step", 100000)

        except AttributeError:
            pass



"""Includes SourceWithPartitionsAlertInfo class"""
from organon.fl.core.helpers import string_helper
from organon.idq.domain.reporting.objects.base_alert_info import BaseAlertInfo


class SourceWithPartitionsAlertInfo(BaseAlertInfo):
    """Alert Info for One DF with Date Info"""

    def to_dict(self):
        """return dict for base alert info"""
        if string_helper.is_null_or_empty(self.full_filter_str):
            return {
                'data_source_name': self.data_source_name,
                'alert': self.alert,
                'run_type': self.run_type.name,
            }
        return {
            'data_source_name': self.data_source_name,
            'filter': self.full_filter_str,
            'alert': self.alert,
            'run_type': self.run_type.name,
        }



"""Includes helper functions for univariate performance calculations"""
from itertools import chain
from typing import List, Dict, Tuple, Optional

import pandas as pd
from sklearn.model_selection import cross_val_score

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.helpers.data_frame_helper import get_column_native_type
from organon.ml.common.enums.target_type import TargetType


def get_reduced_columns_via_performance(data: pd.DataFrame, target_column_name: str, target_type: TargetType,
                                        performance_metric, column_groups: List[List[str]],
                                        univariate_performance_results: Dict[str, float] = None,
                                        random_state=None) -> Tuple[List[str], Optional[Dict[str, float]]]:
    """
      Finds the reduced column using the univariate performance result.

      :return: Returns columns list which reduced and calculated performances for columns
        which are not in initial univariate_performances
      """
    existing_performances = {} if univariate_performance_results is None else univariate_performance_results
    all_cols = set(chain.from_iterable(column_groups))
    cols_to_calculate_performance = [col for col in all_cols if col not in existing_performances]
    reduced_columns = set()
    all_scores = existing_performances
    new_scores = None
    if len(cols_to_calculate_performance) > 0:
        new_scores = get_univariate_performances_for_columns(
            data, cols_to_calculate_performance,
            target_column_name, target_type, performance_metric,
            random_state
        )
        all_scores = {**existing_performances, **new_scores}
    for column_list in column_groups:
        score_results = {col: all_scores[col] for col in column_list if all_scores[col] is not None}
        if len(score_results) > 0:
            best_score_column = max(score_results, key=score_results.get)
            reduced_columns.update([col for col in score_results if col != best_score_column])
    return list(reduced_columns), new_scores


def get_univariate_performances_for_columns(data: pd.DataFrame, columns: List[str],
                                            target_column_name: str, target_type: TargetType,
                                            cross_val_scoring,
                                            random_state: int = None) -> Dict[str, float]:
    """Calculates univariate performances for given column. If column native type is not numeric or string,
    returned scores will be None
    """
    numeric_columns = [col for col in columns if
                       (get_column_native_type(data, col) == ColumnNativeType.Numeric)
                       ]
    string_columns = [col for col in columns if
                      (get_column_native_type(data, col) == ColumnNativeType.String)]
    model = _get_model(target_type, random_state)
    score_results = {}

    for column in columns:
        if column in numeric_columns:
            score = cross_val_score(model, data[[column]], data[target_column_name], cv=3,
                                    scoring=cross_val_scoring).mean()
        elif column in string_columns:
            score = cross_val_score(model, data[[column]].astype('category'),
                                    data[target_column_name], cv=3,
                                    scoring=cross_val_scoring).mean()
        else:
            score = None
        score_results[column] = score
    return score_results


def _get_model(target_type: TargetType, random_state: int):
    """
    Return Model with using target type.
    Parameters
    ----------
    target_type
    random_state

    Returns
    -------
    Returns Model object for LGBM
    """

    if target_type in (TargetType.BINARY, TargetType.MULTICLASS):
        from lightgbm import LGBMClassifier  # pylint: disable=import-outside-toplevel
        model = LGBMClassifier(random_state=random_state)
    else:
        from lightgbm import LGBMRegressor  # pylint: disable=import-outside-toplevel
        model = LGBMRegressor(random_state=random_state)
    return model



"""This module includes helper functions for preprocessing"""
import numpy as np
import pandas as pd


def convert_str_series_to_binary(data: pd.Series, positive_class: str, negative_class: str) -> pd.Series:
    """convert string type series to binary values by given positive and negative classes"""
    if data is None or data.empty:
        raise ValueError("Data is None or empty.")
    if data.nunique(dropna=False) != 2:
        raise ValueError("Target column class count is not equal to 2.")
    if set(data.unique().tolist()) != {positive_class, negative_class}:
        raise ValueError("Target column classes do not match with the given positive and negative classes.")
    res_data = pd.Series(1, index=data.index, dtype=np.uint8)
    res_data.loc[data == negative_class] = 0
    return res_data


def convert_binary_series_to_str(data: pd.Series, positive_class: str, negative_class: str) -> pd.Series:
    """convert binary type series to string values by given positive and negative classes"""
    if data is None or data.empty:
        raise ValueError("Data is None or empty.")
    if data.nunique(dropna=False) != 2:
        raise ValueError("Target column class count is not equal to 2.")
    if set(data.unique().tolist()) != {0, 1}:
        raise ValueError("Target column is not binary.")
    res_data = pd.Series(positive_class, index=data.index, dtype=str)
    res_data.loc[data == 0] = negative_class
    return res_data



"""Includes DetectionInfo class."""
from typing import List


class DetectionInfo:
    """Information generated for a detected object in an image"""

    def __init__(self, object_name: str, confidence: float, box_points: List[int]):
        self.object_name = object_name
        self.confidence = confidence
        self.box_points = box_points



"""Includes UserDateValueDefinition class."""


class UserDateValueDefinition:
    """Date value definition class."""

    def __init__(self):
        self.year: int = None
        self.month: int = None
        self.day: int = None
        self.hour: int = None



"""
This module includes MultiTargetEntityContainerService class.
"""
from typing import List

from organon.afe.dataaccess.services.target_file_record_repository import TargetFileRecordRepository
from organon.afe.domain.modelling.businessobjects.multi_target_entity_container import MultiTargetEntityContainer
from organon.afe.domain.settings.record_source import RecordSource
from organon.afe.domain.settings.target_descriptor import TargetDescriptor
from organon.fl.core.executionutil import parallel_execution_helper


class MultiTargetEntityContainerService:
    """
    Creates a MultiTargetEntityContainer using an AfeModellingSettings instance.
    """

    def __init__(self, target_record_source_list: List[RecordSource], target_descriptor: TargetDescriptor,
                 rejected_target_classes: List[str] = None, is_scoring=False):
        self.__target_record_source_list = target_record_source_list
        self.__target_descriptor = target_descriptor
        self.__container: MultiTargetEntityContainer = None
        self.rejected_target_classes = rejected_target_classes
        self._is_scoring = is_scoring

    @property
    def container(self):
        """Returns value of private attribute '__container'"""
        return self.__container

    def execute(self):
        """
        Sets 'container' value using 'settings'.
        """
        descriptor = self.__target_descriptor

        def inner(record_source):
            repository = self._get_repository(record_source, descriptor, self.rejected_target_classes, self._is_scoring)
            collection = repository.read()
            self.__container.add(record_source, collection)

        self.__container = MultiTargetEntityContainer()

        record_source_list = self.__target_record_source_list
        num_threads = min(len(record_source_list), 25)
        parallel_execution_helper.execute_parallel(record_source_list, inner, num_jobs=num_threads, backend="threading")

    def _get_repository(self, record_source, target_descriptor, rejected_target_classes,
                        is_scoring: bool = False) -> TargetFileRecordRepository:
        # pylint: disable=no-self-use
        return TargetFileRecordRepository(record_source, target_descriptor, rejected_target_classes, is_scoring)



"""Class for Psi Signal Control"""
from abc import ABC
from typing import List, Dict

from organon.idq.core.enums.data_entity_type import DataEntityType
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode
from organon.idq.core.helpers.formatting_helper import FormattingHelper
from organon.idq.domain.businessobjects.dq_calculation_result import DqCalculationResult
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.businessobjects.dq_test_group import DqTestGroup
from organon.idq.domain.controls.base_dq_control import BaseDqControl
from organon.idq.domain.enums.dq_test_group_type import DqTestGroupType
from organon.idq.domain.enums.signal_type import SignalType
from organon.idq.domain.settings.dq_comparison_column_info import DqComparisonColumnInfo


class BasePsiSignalControl(BaseDqControl, ABC):
    """Base class for Psi Signal Control"""

    @classmethod
    def get_test_group_type(cls) -> DqTestGroupType:
        return DqTestGroupType.DISTRIBUTIONAL_COMPARISONS

    def check_maximum_psi(self, col_name: str, result_code: DqComparisonResultCode,
                          results: List[DqComparisonResult], max_psi: float, property_key: str):
        """Result for MaximumPsi"""
        result = DqComparisonResult(
            data_entity=DataEntityType.COLUMN,
            data_entity_name=col_name,
            test_group=self.get_test_group_type(),
            result_code=result_code,
            property_code=property_key,
            message=f"Distribution is unstable for the nominal column {col_name}")
        result.property_key_numeric = FormattingHelper.to_nullable(max_psi)
        results.append(result)

    def check_maximum_psi_control_set(self, col_name: str, result_code: DqComparisonResultCode,
                                      results: List[DqComparisonResult], maximum_psi_sample: str, property_key: str):
        """Result for MaximumPsiControlSet"""
        result = DqComparisonResult(
            data_entity=DataEntityType.COLUMN,
            data_entity_name=col_name,
            test_group=self.get_test_group_type(),
            result_code=result_code,
            property_code=property_key,
            message=f"Distribution is unstable for the nominal column {col_name}")
        result.property_key_nominal = maximum_psi_sample
        results.append(result)

    def check_maximum_psi_category(self, col_name: str, result_code: DqComparisonResultCode,
                                   results: List[DqComparisonResult], maximum_psi_category: str, property_code: str):
        """Result for MaximumPsiCategory"""
        result = DqComparisonResult(
            data_entity=DataEntityType.COLUMN,
            data_entity_name=col_name,
            test_group=self.get_test_group_type(),
            result_code=result_code,
            property_code=property_code,
            message=f"Distribution is unstable for the nominal column {col_name}")
        result.property_key_nominal = maximum_psi_category
        results.append(result)

    def check_test_frequency(self, col_name: str, result_code: DqComparisonResultCode,
                             results: List[DqComparisonResult], test_frequencies: Dict[object, float],
                             property_key: str):
        "Result for TestFrequencies"
        if test_frequencies is not None:
            for test_frequency_key, test_frequency_value in test_frequencies.items():
                result = DqComparisonResult(
                    data_entity=DataEntityType.COLUMN,
                    data_entity_name=col_name,
                    test_group=self.get_test_group_type(),
                    result_code=result_code,
                    property_code=property_key,
                    message=f"Distribution is unstable for the nominal column {col_name}")
                result.property_key_nominal = test_frequency_key
                result.property_key_numeric = FormattingHelper.to_nullable(test_frequency_value)
                results.append(result)

    def check_max_psi_control_frequency(self, col_name: str, result_code: DqComparisonResultCode,
                                        results: List[DqComparisonResult],
                                        maximum_psi_control_set_frequencies: Dict[str, float], property_key: str):
        """Test for MaximumPsiControlFrequencies"""
        if maximum_psi_control_set_frequencies is not None:
            for control_set_frequency_key, control_set_frequency_value \
                    in maximum_psi_control_set_frequencies.items():
                result = DqComparisonResult(
                    data_entity=DataEntityType.COLUMN,
                    data_entity_name=col_name,
                    test_group=self.get_test_group_type(),
                    result_code=result_code,
                    property_code=property_key,
                    message=f"Distribution is unstable for the nominal column {col_name}")
                result.property_key_nominal = control_set_frequency_key
                result.property_key_numeric = FormattingHelper.to_nullable(control_set_frequency_value)
                results.append(result)

    def check_psi_per_control_set(self, col_name: str, result_code: DqComparisonResultCode,
                                  results: List[DqComparisonResult],
                                  psi_per_control_set: Dict[str, float], property_code: str):
        """Test for PsiPerControlSet"""
        if len(psi_per_control_set) > 0:
            for control_set_psi_key, control_set_psi_value in psi_per_control_set.items():
                result = DqComparisonResult(
                    data_entity=DataEntityType.COLUMN,
                    data_entity_name=col_name,
                    test_group=self.get_test_group_type(),
                    result_code=result_code,
                    property_code=property_code,
                    message=f"Distribution is unstable for the nominal column {col_name}")
                result.property_key_nominal = str(control_set_psi_key)
                result.property_key_numeric = FormattingHelper.to_nullable(control_set_psi_value)
                results.append(result)

    def get_signal_type(self, max_psi: float) -> SignalType:
        """Return Signal Type """
        psi_threshold_yellow = self.comp_context.comparison_parameters.psi_threshold_green
        psi_threshold_red = self.comp_context.comparison_parameters.psi_threshold_yellow
        if max_psi < psi_threshold_yellow:
            psi_signal = SignalType.GREEN
        elif max_psi < psi_threshold_red:
            psi_signal = SignalType.YELLOW
        else:
            psi_signal = SignalType.RED
        return psi_signal

    def _get_filtered_columns(self, col: DqComparisonColumnInfo, test_group_info: DqTestGroup) -> \
            List[DqCalculationResult]:
        effective_bmh = min(test_group_info.test_bmh, col.benchmark_horizon)
        max_index = max(len(self.comp_context.control_results) - effective_bmh, 0)
        filtered_controls = list(reversed(self.comp_context.control_results[max_index:]))
        return filtered_controls



"""
This module includes AfeScoringSettings class.
"""

from organon.afe.domain.reporting.afe_model_output import AfeModelOutput
from organon.afe.domain.settings.base_afe_scoring_settings import BaseAfeScoringSettings


class AfeScoringSettings(BaseAfeScoringSettings[AfeModelOutput]):
    """
    Scoring settings for AFE
    """
    ATTR_DICT = {
        "model_output": AfeModelOutput
    }
    ATTR_DICT.update(BaseAfeScoringSettings.ATTR_DICT)



"""
This module includes enum class AfeTargetColumnType.
"""
from enum import Enum


# pylint: disable=invalid-name
class AfeTargetColumnType(Enum):
    """
    AfeTargetColumnType
    """
    Binary = 0
    Scalar = 1
    MultiClass = 3



"""
This module includes AfeOutputSettings class.
"""
from organon.afe.domain.common.reader_helper import get_values_from_kwargs


class AfeOutputSettings:
    """
    Settings for output of Automated Feature Extraction
    """
    ATTR_DICT = {
        "output_folder": str,
        "output_prefix": str,
        "feature_name_prefix": str,
        "enable_write_output": bool,
        "enable_feature_lookup_output_to_csv": bool,
        "return_all_afe_columns": bool,
        "enable_all_feature_lookup_output_to_csv": bool
    }

    def __init__(self, **kwargs):
        self.output_folder: str = None
        self.output_prefix: str = None
        self.feature_name_prefix: str = None

        self.enable_feature_lookup_output_to_csv: bool = None
        self.enable_write_output: bool = None
        self.return_all_afe_columns: bool = None
        self.enable_all_feature_lookup_output_to_csv = None

        get_values_from_kwargs(self, self.ATTR_DICT, kwargs)

    def get_feature_table_name(self) -> str:
        """
        :return: name for feature table
        """
        return f"{self.output_prefix}_FEATURE_DVL"

    def get_feature_lookup_table_name(self) -> str:
        """
        :return: name for feature lookup table
        """
        return f"{self.output_prefix}_LOOKUP"

    def get_all_features_lookup_table_name(self) -> str:
        """
        :return: name for all lookup table
        """
        return f"{self.output_prefix}_ALL_LOOKUP"

    def get_sql_file(self) -> str:
        """
        :return: sql file name
        """
        return f"{self.output_prefix}_SQL"

    def get_distinct_entities_file(self) -> str:
        """
        :return: distinct entity table name
        """
        return f"{self.output_prefix}_DISTINCT_ENTITY_TABLE"



"""Includes DqCalculationOutput class."""
from organon.idq.domain.businessobjects.dq_calculation_result import DqCalculationResult
from organon.idq.domain.settings.abstractions.dq_base_calculation_parameters import DqBaseCalculationParameters


class DqCalculationOutput:
    """Stores dq calculation results with its parameters and extra info of execution"""

    def __init__(self):
        self.calculation_parameters: DqBaseCalculationParameters = None
        self.calculation_result: DqCalculationResult = None



"""Includes UserHPOptimizationSettings class."""
from dataclasses import dataclass
from typing import List, Dict, Optional

import pandas as pd


@dataclass
class UserHPOptimizationSettings:
    """User settings class for HPOptimizer"""
    train_data: pd.DataFrame
    target_data: pd.Series
    search_method: str
    modellers: List[str] = None
    modeller_params: List[Dict[str, list]] = None
    cv_fold: int = None
    scoring_metrics: Optional[List[str]] = None
    test_data: pd.DataFrame = None
    test_target_data: pd.Series = None
    search_params: dict = None



"""todo"""


class StatsOutputService:
    """todo"""

    def write_table_stats(self):
        """todo"""

    def write_sample_stats(self):
        """todo"""

    def write_population_stats(self):
        """todo"""



"""Includes BaseAlertInfo class"""
import abc

from organon.fl.core.helpers import string_helper
from organon.idq.domain.enums.dq_run_type import DqRunType


class BaseAlertInfo(metaclass=abc.ABCMeta):
    """Class for base alert info"""
    def __init__(self):
        self.data_source_name: str = None
        self.full_filter_str: str = None
        self.alert: str = None
        self.run_type: DqRunType = None

    def to_dict(self):
        """return dict for base alert info"""
        if string_helper.is_null_or_empty(self.full_filter_str):
            return {
                'data_source_name': self.data_source_name,
                'alert': self.alert,
                'run_type': self.run_type.name
            }
        return {
            'data_source_name': self.data_source_name,
            'filter': self.full_filter_str,
            'alert': self.alert,
            'run_type': self.run_type.name
        }



"""Includes UserInputValidationService class."""
from typing import Dict

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.idq.domain.services.base_user_input_validation_service import BaseUserInputValidationService
from organon.idq.services.user_settings.dq_user_input import DqUserInput


class UserInputValidationService(BaseUserInputValidationService[DqUserInput]):
    """Service for validating settings in DqUserInput"""

    def __init__(self, user_input: DqUserInput, column_native_types: Dict[str, ColumnNativeType]):
        super().__init__(user_input)
        self.column_native_types = column_native_types

    def validate(self):
        """Executes validation"""
        if self.user_input.calc_params is None:
            raise KnownException("Calculation settings not given")

        super().validate()

        self._validate_calculation_parameters(0, self.user_input.calc_params, self.column_native_types)
        self._validate_included_excluded_columns([list(self.column_native_types.keys())])
        self._validate_comparison_column_info(list(self.column_native_types.keys()))
        self._validate_column_metadata_list(list(self.column_native_types.keys()))



"""
This module includes RuntimeStatistics class.
"""
from datetime import datetime
from typing import Dict


class RuntimeStatistics:
    """Runtime statistics."""
    def __init__(self):
        self.execution_time: float = None
        self.user_cpu_time: float = None
        self.system_cpu_time: float = None
        self.memory_usage: Dict[datetime, float] = {}
        self.cpu_usage: Dict[datetime, float] = {}
        self.event_times: Dict[str, datetime] = {}



"""Includes MulticlassReport class."""
from typing import Dict, Tuple

import pandas as pd

from organon.ml.reporting.domain.objects.base_report import BaseReport


class MulticlassReport(BaseReport):
    """Report for multiclass classification"""

    def __init__(self):
        self.scores: pd.DataFrame = None
        self.confusion_matrices: Dict[Tuple, pd.DataFrame] = None
        self.classification_reports: Dict[Tuple, pd.DataFrame] = None



"""
This module keeps the functions used in file io operations.
"""
import io
import logging
import os
import pkgutil
from typing import List

from organon.fl.core.fileoperations import directory_helper
from organon.fl.core.fileoperations.file_read_options import FileReadOptions


def read_all_lines(file_read_options: FileReadOptions) -> List[str]:
    """
    Reads the file line by line, and returns a list of strings(lines) of the file.
    :param file_read_options: FileReadOptions object
    :return: a list of strings(lines)
    """
    read_bytes = None
    lines: List[str] = []
    read_str: str
    with io.open(file_read_options.file_full_path, "rb") as read_bytes:
        read_str = read_bytes.readline().decode("utf-8")
        if read_str is None:
            return lines
        read_str = read_str.replace("\uFEFF", "")
        lines.append(read_str)
        read_str = read_bytes.readline().decode("utf-8")
        while read_str != '':
            lines.append(read_str)
            read_str = read_bytes.readline().decode("utf-8")

    return lines


def read_all_text_from_file(path: str) -> str:
    """
    Reads all text of the file.
    :param path: file path
    :return: all text of the file
    """
    with open(path, 'r', encoding="utf8") as file:
        read_file = file.read()
    return read_file


def read_all_text_from_resource_file(resource_name: str) -> str:
    """
    Reads all text of a resource file.
    :param resource_name: file path
    :return: all text of the file
    """
    result: str = ""
    try:
        result = pkgutil.get_data(__name__, resource_name)
    except (ModuleNotFoundError, ValueError, ImportError, NotImplementedError) as ex:
        logging.error("An error occurred while reading the resource file.")
        logging.error(ex)
    return result


def write_to_file(file_path: str, file_detail: str, append: bool = False, ensure_path: bool = False):
    """
    Writes a string to the file.
    :param file_path: file path
    :param file_detail: string to write
    :param append: append to end of file
    :param ensure_path: create path if not exists
    :return:
    """
    mode = "w+b" if not append else "a+b"
    if ensure_path:
        directory_helper.create(get_directory_name(file_path))
    with open(file_path, mode) as out:
        out.write(bytearray(file_detail, 'utf-8'))


def check_file_exists(file_path: str) -> bool:
    """
    Checks if the file exists in the parameter path.
    :param file_path: file path
    :return: bool checks whether the file exists in the parameter path
    """
    return os.path.isfile(file_path) and not os.path.isdir(file_path)


def get_directory_name(file_path: str) -> str:
    """Returns directory name of given file path"""
    return os.path.dirname(file_path)



"""Module Description"""
import numpy as np
# pylint: disable=no-name-in-module
from organon.fl.mathematics.sweep.pyx_files.sweep_helper import get_correlation_matrix_c


class CovarianceInfo:
    """Class Description"""
    @property
    def correlation_matrix(self):
        """Get the correlation matrix"""
        return self._correlation_matrix

    @property
    def covariance_matrix(self):
        """Get the covariance matrix"""
        return self._covariance_matrix

    @property
    def means(self):
        """Get the means"""
        return self._means

    @property
    def total_case_weight(self):
        """Get the total case weight"""
        return self._total_case_weight

    def __init__(self, covariance_matrix: np.ndarray, total_case_weight: int):
        CovarianceInfo._validate(covariance_matrix, total_case_weight)
        self._covariance_matrix = covariance_matrix
        # self._correlation_matrix = CovarianceInfo.get_correlation_matrix(covariance_matrix)
        self._correlation_matrix = get_correlation_matrix_c(covariance_matrix, 4)

        self._means = np.empty(shape=(covariance_matrix.shape[0],), dtype=self._covariance_matrix.dtype)
        self._means.fill(float('nan'))

        self._total_case_weight = total_case_weight

    @staticmethod
    def _validate(covariance_matrix: np.ndarray, total_case_weight: int):
        """Validates input elements"""
        if covariance_matrix is None:
            raise ValueError("Covariance matrix can not be null in CovarianceInfo constructor.")

        if covariance_matrix.shape[0] != covariance_matrix.shape[1]:
            raise ValueError("Covariance matrix must be a square matrix.")

        if total_case_weight <= 0:
            raise ValueError("TotalCaseWight must be positive in CovarianceInfo constructor.")

    def update_correlation_matrix(self):
        """Updates correlation matrix according to covariance matrix."""
        self._correlation_matrix = CovarianceInfo.get_correlation_matrix(self._covariance_matrix)

    @staticmethod
    def get_correlation_matrix(covariance_matrix: np.ndarray) -> np.ndarray:
        """Gets correlation matrix from covariance matrix"""
        matrix = covariance_matrix
        rows = matrix.shape[0]
        cols = matrix.shape[1]

        if rows != cols:
            raise ValueError("Covariance shape doesn't fit for getting correlation matrix."
                             f" Rows = {rows}, Cols = {cols}")

        sigma = np.empty(shape=(rows, cols), dtype=matrix.dtype)

        for i in range(rows):
            nan_i = np.isnan(matrix[i, i])
            zero_i = matrix[i, i] <= 0

            sigma[i, i] = float('nan') if nan_i else 1.0

            for j in range(i, cols):
                nan_j = np.isnan(matrix[j, j])
                zero_j = matrix[j, j] <= 0

                if np.isnan(matrix[i, j]) or nan_i or nan_j:
                    sigma[i, j] = float('nan')

                elif zero_i or zero_j:
                    sigma[i, j] = 0.0

                else:
                    sigma[i, j] = matrix[i, j] / np.sqrt(matrix[i, i] * matrix[j, j])

                sigma[j, i] = sigma[i, j]

        return sigma

    def get_non_zero_variance_attributes(self) -> set:
        """Returns non zero variance attributes"""
        indices = set()

        for i in range(self._covariance_matrix.shape[0]):
            variance = self._covariance_matrix[i, i]

            if np.isnan(variance) or variance <= 0:
                continue
            indices.add(i)

        return indices

    def get_variances(self, linear_form) -> float:
        """Returns variances for matrix * linear_form(vector) multiplication"""
        rows = self._covariance_matrix.shape[0]

        if any(key >= rows for key in linear_form):
            raise IndexError("At least one variable in the linear form does not correspond to covariance matrix"
                             " variables")

        variance = 0.0

        for i in linear_form:
            beta_i = linear_form[i]
            for j in linear_form:
                beta_j = linear_form[j]
                variance += beta_i * beta_j * self._covariance_matrix[i, j]

        return variance

    def get_covariances(self, linear_form, target) -> float:
        """Calculates covariance of matrix vector product"""
        rows = self._covariance_matrix.shape[0]

        if any(key >= rows for key in linear_form):
            raise IndexError("At least one variable in the linear form does not correspond to covariance matrix"
                             " variables")
        if target >= rows:
            raise IndexError("The target variable does not correspond to any covariance matrix variable")

        variance = 0.0
        for i in linear_form:
            beta_i = linear_form[i]
            variance += beta_i * self._covariance_matrix[i, target]

        return variance



"""Includes BaseRegressor class."""
import abc
from typing import Union

import pandas as pd
from sklearn.metrics import r2_score

from organon.ml.modelling.algorithms.core.abstractions.base_modeller import BaseModeller
from organon.ml.modelling.algorithms.core.enums.modeller_type import ModellerType


class BaseRegressor(BaseModeller, metaclass=abc.ABCMeta):
    """Base class for regressors"""
    _estimator_type = "regressor"  # REQUIRED FOR SKLEARN GridSearch

    def _score(self, train_data: pd.DataFrame, target_data: Union[pd.DataFrame, pd.Series],
               sample_weight=None) -> float:
        y_pred = self.predict(train_data)
        return r2_score(target_data, y_pred, sample_weight=sample_weight)

    @property
    def modeller_type(self) -> ModellerType:
        return ModellerType.REGRESSOR



"""Includes BaseClassifier class."""
import abc
from typing import List, Union

import pandas as pd
from sklearn.metrics import accuracy_score

from organon.ml.modelling.algorithms.core.abstractions.base_modeller import BaseModeller
from organon.ml.modelling.algorithms.core.enums.modeller_type import ModellerType


class BaseClassifier(BaseModeller, metaclass=abc.ABCMeta):
    """Base class for classifiers"""
    _estimator_type = "classifier"  # REQUIRED FOR SKLEARN GridSearch

    def predict_proba(self, data: pd.DataFrame) -> pd.DataFrame:
        """Predict score as probability"""
        self._check_is_fitted()
        return self._predict_proba(data)

    @abc.abstractmethod
    def _predict_proba(self, data: pd.DataFrame) -> pd.DataFrame:
        raise NotImplementedError

    @property
    def classes_(self):
        """Returns target class values in the same order as in return value of predict_proba"""
        # This property should return a valid value only after fit is called
        if not self.is_fitted():
            raise AttributeError("Modeller not fitted yet")
        return self._get_classes()

    @abc.abstractmethod
    def _get_classes(self) -> List[str]:
        """Returns target class values in the same order as in return value of predict_proba"""

    def _score(self, train_data: pd.DataFrame, target_data: Union[pd.DataFrame, pd.Series],
               sample_weight=None) -> float:
        return self.get_score_from_predictions(self.predict(train_data), target_data)

    @classmethod
    def get_score_from_predictions(cls, predictions: pd.DataFrame, actual: pd.DataFrame, sample_weight=None):
        """Calculates score from predictions and actual data"""
        return accuracy_score(actual, predictions, sample_weight=sample_weight)

    @property
    def modeller_type(self) -> ModellerType:
        return ModellerType.CLASSIFIER



"""Module for Transformation"""
from dataclasses import dataclass


@dataclass
class Transformation:
    """Class Description"""
    _default_output_value: int
    _variable_importance: int
    variable_importance_raw: int
    is_stable: bool

    @staticmethod
    def evaluate(inp):
        """Function Description"""
        return inp



"""Includes CrossValidationService class"""
import numpy as np
import pandas as pd
from sklearn.model_selection import StratifiedKFold, PredefinedSplit, cross_val_score
from sklearn.preprocessing import KBinsDiscretizer

from organon.fl.mathematics.constants import DOUBLE_MIN

from organon.ml.modelling.algorithms.core.enums.modeller_type import ModellerType
from organon.ml.modelling.model_selection.domain.objects.cross_validation_output import CrossValidationOutput
from organon.ml.modelling.model_selection.settings.objects.cross_validation_settings import CrossValidationSettings

DEFAULT_CV_COUNT = 3
DEFAULT_BIN_COUNT = 10


class CrossValidationService:
    """Service for cross validation"""

    @classmethod
    def execute(cls, settings: CrossValidationSettings) -> CrossValidationOutput:
        """Runs multiple modellers with cross validation using same folds and report scores for every run"""
        modellers = settings.modellers
        modelling_type = modellers[0].modeller_type

        pds = cls.get_predefined_split(settings.train_data, settings.target_data, modelling_type, settings.cv_count,
                                       bin_count=settings.bin_count)

        output = CrossValidationOutput()
        output.fold_scores = []
        output.mean_scores = []
        best_score = DOUBLE_MIN
        best_modeller = None
        for modeller in modellers:  # modellers should be fitted in order
            res = cross_val_score(modeller, settings.train_data, settings.target_data, cv=pds)
            output.fold_scores.append(list(res))
            mean_score = np.mean(res)
            output.mean_scores.append(mean_score)
            if mean_score > best_score:
                best_score = mean_score
                best_modeller = modeller
        output.best_modeller = best_modeller
        if settings.return_test_fold:
            output.test_fold = pds.test_fold
        return output

    @classmethod
    def get_predefined_split(cls, train_data: pd.DataFrame, target_data: pd.Series, modeller_type: ModellerType,
                             cv_count: int = DEFAULT_CV_COUNT, bin_count: int = DEFAULT_BIN_COUNT):
        """Returns a PredefinedSplit instance for given train and target data"""
        if modeller_type == ModellerType.REGRESSOR:
            if bin_count is None:
                raise ValueError("bin_count should not be None for regression")
            target_data = cls._get_discretized(target_data, bin_count)
        return cls.get_predefined_split_for_stratified_target(train_data, target_data, cv_count)

    @classmethod
    def get_predefined_split_for_stratified_target(cls, train_data: pd.DataFrame, target_data: pd.Series,
                                                   cv_count: int = None):
        """Returns a PredefinedSplit instance for given train and target data.
        Should only be used for classification modelling."""
        fold = StratifiedKFold(n_splits=cv_count, shuffle=True)
        # StratifiedKFold uses target as the strata to generate stratified splits
        split_generator = fold.split(train_data, target_data)
        test_fold = np.full(len(train_data), -1)
        counter = 0
        for _, test_indices in split_generator:
            test_fold[test_indices] = counter
            counter += 1
        return PredefinedSplit(test_fold)

    @classmethod
    def _get_discretized(cls, target_data: pd.Series, bin_count: int = DEFAULT_BIN_COUNT) -> np.ndarray:
        discretizer = KBinsDiscretizer(n_bins=bin_count, encode="ordinal", strategy="quantile")
        target_data = discretizer.fit_transform(target_data.to_numpy().reshape(-1, 1))[:, 0]
        return target_data



"""Includes SamplingService class."""
from typing import List, Dict, Any, Tuple, TypeVar

import pandas as pd

from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.ml.common.helpers import df_ops_helper
from organon.ml.sampling.domain.objects.sampling_output import SamplingOutput
from organon.ml.sampling.domain.services.sampling_preprocessing_service import SamplingPreprocessingService
from organon.ml.sampling.settings.enums.sampling_strategy import SamplingStrategy
from organon.ml.sampling.settings.objects.sampling_settings import SamplingSettings

TargetClassValue = TypeVar("TargetClassValue")


# THIS CLASS IS OPTIMIZED FOR LOW PEAK MEMORY USAGE AND EXECUTION TIME.
# MEASURE PEAK MEMORY USAGE AND EXECUTION TIMES BEFORE AND AFTER DEVELOPMENT.

class SamplingService:
    """Sampling service"""

    @classmethod
    def sample(cls, sampling_settings: SamplingSettings) -> SamplingOutput:
        """Samples data according to given settings and returns sampled train data (and test data optionally)."""
        strata_cols = sampling_settings.strata_columns
        data = SamplingPreprocessingService.discretize(sampling_settings.data, strata_cols)

        if sampling_settings.split_test_data:
            x_train, x_test = df_ops_helper.get_train_test_split(data, sampling_settings.test_split_ratio,
                                                                 strata_columns=strata_cols)
        else:
            x_train = data
            x_test = None

        if sampling_settings.sampling_strategy == SamplingStrategy.UNDERSAMPLING:
            if sampling_settings.data_sample_ratio is not None:
                x_train = cls._get_random_sampled(x_train, sampling_settings.data_sample_ratio, strata_cols)

            x_train = cls._get_undersampled(x_train, sampling_settings.target_column_name,
                                            sampling_settings.sampling_ratio,
                                            strata_cols)
        elif sampling_settings.sampling_strategy == SamplingStrategy.OVERSAMPLING:
            if sampling_settings.data_sample_ratio is not None:
                x_train = cls._get_random_sampled(x_train, sampling_settings.data_sample_ratio, strata_cols)

            x_train = cls._get_oversampled(x_train, sampling_settings.target_column_name,
                                           sampling_settings.sampling_ratio,
                                           strata_cols)
        elif sampling_settings.sampling_strategy == SamplingStrategy.RANDOM_SAMPLING:
            x_train = cls._get_random_sampled(x_train, sampling_settings.sampling_ratio, strata_cols)

        output = SamplingOutput()
        output.train_data = x_train
        output.test_data = x_test
        return output

    @classmethod
    def _get_undersampled(cls, data: pd.DataFrame, target_column: str, sampling_ratio: float,
                          strata_columns: List[str]):
        target_value_counts: dict = data[target_column].value_counts(dropna=False).to_dict()
        target_classes = list(target_value_counts.keys())
        if len(target_classes) == 2:
            return cls._get_undersampled_for_binary_target(data, target_value_counts, target_column, sampling_ratio,
                                                           strata_columns)
        return cls._get_undersampled_for_multiclass_target(data, target_column, target_classes, target_value_counts,
                                                           strata_columns)

    @classmethod
    def _get_oversampled(cls, data: pd.DataFrame, target_column: str, sampling_ratio: float,
                         strata_columns: List[str]):
        target_value_counts: dict = data[target_column].value_counts(dropna=False).to_dict()
        target_classes = list(target_value_counts.keys())

        if len(target_classes) == 2:
            return cls._get_oversampled_for_binary_target(data, target_value_counts, target_column, sampling_ratio,
                                                          strata_columns)
        return cls._get_oversampled_for_multiclass_target(data, target_column, target_classes, target_value_counts,
                                                          strata_columns)

    @classmethod
    def _get_undersampled_for_binary_target(cls, data: pd.DataFrame, target_value_counts: Dict[Any, int],
                                            target_column: str,
                                            sampling_ratio: float, strata_columns: List[str]):
        min_freq_class, min_freq = cls._get_min_frequency_and_class(target_value_counts)
        expected_length_for_other = int(min_freq / sampling_ratio) - min_freq
        min_freq_class_rows = data[target_column] == min_freq_class

        df_min_class_indices = data.index[min_freq_class_rows]
        len_data = sum(target_value_counts.values())
        len_other = len_data - min_freq
        if expected_length_for_other > len_other:
            raise KnownException("Undersampling ratio cannot be less than the ratio of minimum sized class")
        if expected_length_for_other == len_other:
            return data
        df_other_indices = cls._get_sample_indices(cls._get_strata_columns(data[~min_freq_class_rows], strata_columns),
                                                   strata_columns,
                                                   expected_length_for_other / len_other)
        return data.loc[df_min_class_indices.tolist() + df_other_indices.tolist()]

    @classmethod
    def _get_undersampled_for_multiclass_target(cls, data: pd.DataFrame, target_column: str, target_classes: list,
                                                target_value_counts: Dict[Any, int], strata_columns: List[str]):
        min_freq_class, min_freq = cls._get_min_frequency_and_class(target_value_counts)
        min_freq_class_rows = data[target_column] == min_freq_class
        all_indices = data.index[min_freq_class_rows].tolist()
        for target_class in target_classes:
            if target_class == min_freq_class:
                continue
            target_class_size = target_value_counts[target_class]
            target_class_indices = cls._get_sample_indices(
                cls._get_strata_columns(data[data[target_column] == target_class], strata_columns),
                strata_columns, min_freq / target_class_size)
            all_indices.extend(target_class_indices)
        return data.loc[all_indices]

    @classmethod
    def _get_min_frequency_and_class(cls, target_value_counts: Dict[TargetClassValue,
                                                                    int]) -> Tuple[TargetClassValue, int]:
        min_freq = min(target_value_counts.values())
        min_freq_class = next(val for val, frequency in target_value_counts.items() if frequency == min_freq)
        return min_freq_class, min_freq

    @classmethod
    def _get_max_frequency_and_class(cls, target_value_counts: Dict[TargetClassValue,
                                                                    int]) -> Tuple[TargetClassValue, int]:
        max_freq = max(target_value_counts.values())
        max_freq_class = next(val for val, frequency in target_value_counts.items() if frequency == max_freq)
        return max_freq_class, max_freq

    @classmethod
    def _get_oversampled_for_binary_target(cls, data: pd.DataFrame, target_value_counts: Dict[Any, int],
                                           target_column: str, sampling_ratio: float, strata_columns: List[str]):
        max_freq_class, max_freq = cls._get_max_frequency_and_class(target_value_counts)
        max_freq_class_rows = data[target_column] == max_freq_class

        data_len = sum(target_value_counts.values())
        len_other = data_len - max_freq
        expected_length_for_other_class = int(max_freq / (1 - sampling_ratio) - max_freq)
        expected_ratio = expected_length_for_other_class / len_other

        df_max_class = data.index[max_freq_class_rows]
        if expected_ratio < 1:
            raise KnownException("Oversampling ratio cannot be less than the ratio of minimum sized class")
        if expected_ratio == 1:
            return data
        indices = cls._get_sample_indices(cls._get_strata_columns(data[~max_freq_class_rows], strata_columns),
                                          strata_columns, frac=expected_ratio, replace=True)
        return data.loc[df_max_class.tolist() + indices.tolist()]

    @classmethod
    def _get_oversampled_for_multiclass_target(cls, data: pd.DataFrame, target_column: str, target_classes: list,
                                               target_value_counts: Dict[Any, int], strata_columns: List[str]):
        max_freq_class, max_freq = cls._get_max_frequency_and_class(target_value_counts)
        max_freq_class_rows = data[target_column] == max_freq_class
        new_df_indices = data.index[max_freq_class_rows].tolist()
        for target_class in target_classes:
            if target_class == max_freq_class:
                continue
            len_df_class = target_value_counts[target_class]
            indices = cls._get_sample_indices(cls._get_strata_columns(data[data[target_column] == target_class],
                                                                      strata_columns), strata_columns,
                                              frac=max_freq / len_df_class, replace=True)
            new_df_indices.extend(indices)
        return data.loc[new_df_indices]

    @classmethod
    def _get_random_sampled(cls, data: pd.DataFrame, sampling_ratio: float,
                            strata_columns: List[str]):
        indices = cls._get_sample_indices(cls._get_strata_columns(data, strata_columns), strata_columns,
                                          sampling_ratio)
        return data.loc[indices]

    @classmethod
    def _get_sample_indices(cls, data: pd.DataFrame, strata_columns: List[str], frac: float, replace=False):
        return df_ops_helper.get_sample_indices(data, frac, strata_columns=strata_columns, replace=replace)

    @classmethod
    def _get_strata_columns(cls, data: pd.DataFrame, strata_columns: List[str]):
        if strata_columns is None:
            return None
        return data[strata_columns]



"""Includes BaseFeatureSelecter class."""
import abc

import pandas as pd


class BaseFeatureSelecter(metaclass=abc.ABCMeta):
    """Base user service for feature selection"""

    def __init__(self, data: pd.DataFrame):
        self.data = data

    def run(self):
        """Run selecter and return output"""
        return self._run()

    @abc.abstractmethod
    def _run(self):
        raise NotImplementedError



"""Includes DqDataSource class"""
from typing import Optional

import pandas as pd


class DqDataSource:
    """DqDataSource attributes"""

    def __init__(self):
        self.sampled_data: Optional[pd.DataFrame] = None
        self.full_data_row_count: int = None



"""This module includes helper function for creating dashboards with plotly."""
import math
import textwrap
from typing import List

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.axes import Axes, BarContainer
from matplotlib.figure import Figure
from matplotlib.lines import Line2D
from matplotlib.table import Table
from matplotlib.widgets import Button

ORG_COLOR_HEX = "#2D1A44"


# ax kıslatması bir convention
# pylint: disable=invalid-name

def get_bar_chart(x_vals: List[str], y_vals: List[int or float], label="", ax: Axes = None) -> BarContainer:
    """
    Creates a bar chart with given bar names and values
    :param x_vals: Names of bars
    :param y_vals: Values of bars
    :param label: Bar chart title
    :param ax: Axes object where bar chart will be added
    :return: Matplotlib bar container with given values
    """
    ax = ax or plt.gca()
    bar_container = ax.bar(x_vals, y_vals, label=label)
    for index_elem in enumerate(ax.get_xticks()):
        i = index_elem[0]
        ax.text(i, y_vals[i] / 2, y_vals[i], horizontalalignment='center',
                verticalalignment='center', color='white')
    return bar_container


def get_line_scatter_plot(x_vals: List[int or float], y_vals: List[int or float], title: str = None,
                          ax: Axes = None, color=None) -> Line2D:
    """
    Creates a line plot where data points are marked
    :param x_vals: X axis values of data points
    :param y_vals: Y axis values of data points
    :param title: Title of plot
    :param ax: Axes object where bar chart will be added
    :param color: color of plot
    :return: Line2D object with given values
    """
    ax = ax or plt.gca()
    if color is None:
        plot = ax.plot(x_vals, y_vals, label=title, marker=".")
    else:
        plot = ax.plot(x_vals, y_vals, label=title, marker=".", color=color)
    return plot


class TablePagination:
    """Converts a matplotlib table to a paginated table"""

    def __init__(self, ax: Axes, table: Table, data=List[List[str or int or float]], rows_per_page=8):
        self.current_page_index = 0
        self.table = table
        self.ax = ax or plt.gca()
        self.data = data
        self.num_rows = len(data)
        self.rows_per_page = rows_per_page
        self.current_data = []
        self.num_pages = math.ceil(self.num_rows / self.rows_per_page)
        self.cells = table.get_celld()
        self.pages_text = ax.annotate(f"Page 1/{self.num_pages}", xy=(0, 0),
                                      xycoords='axes fraction', fontsize=self.cells[(0, 0)].get_fontsize(),
                                      weight='bold', )
        self.btn_prev = None
        self.btn_next = None

    def next(self, event):  # pylint: disable=unused-argument
        """This method should be called for passing to next page of table.
        Can be used as a callback in button on_clicked methods."""
        if self.current_page_index < self.num_pages - 1:
            self.current_page_index += 1
            self._update_rows()
            self._update_texts()
            plt.draw()

    def prev(self, event):  # pylint: disable=unused-argument
        """This method should be called for returning to previous page of table.
        Can be used as a callback in button on_clicked methods."""
        if self.current_page_index > 0:
            self.current_page_index -= 1
            self._update_rows()
            self._update_texts()
            plt.draw()

    def _update_texts(self):
        """Updates texts in table cells according to values in current page"""
        col_num = len(self.data[0])
        for i in range(1, len(self.current_data) + 1):
            for j in range(col_num):
                txt = _wrap_text_for_table(self.current_data[i - 1][j])
                self.cells[(i, j)].set_text_props(text=txt)
        for i in range(len(self.current_data) + 1, self.rows_per_page + 1):
            for j in range(col_num):
                self.cells[(i, j)].set_text_props(text="")
        self.pages_text.set_text(f"Page {self.current_page_index + 1}/{self.num_pages}")

    def _update_rows(self):
        """Sets currently shown data according to page number"""
        lower = self.current_page_index * self.rows_per_page
        upper = min((self.current_page_index + 1) * self.rows_per_page, self.num_rows)
        self.current_data = self.data[lower:upper]

    def add_buttons(self, fig: Figure, btn_width=0.05, btn_height=0.035):
        """Adds prev/next buttons under the table"""
        bbox = self.ax.get_position()

        prev_left = bbox.x1 - 3 * btn_width
        prev_bottom = bbox.y0 - 1 * btn_height
        axprev = fig.add_axes([prev_left, prev_bottom, btn_width, btn_height])

        next_left = prev_left + 1.3 * btn_width
        next_bottom = prev_bottom
        axnext = fig.add_axes([next_left, next_bottom, btn_width, btn_height])

        self.btn_prev = Button(axprev, "Previous")
        self.btn_next = Button(axnext, "Next")

        self.btn_prev.on_clicked(self.prev)
        self.btn_next.on_clicked(self.next)


def get_data_table_plot(data_frame: pd.DataFrame, ax: Axes = None, rows_per_page=10):
    """
    Creates a paginated data table from given data frame
    :param data_frame: Data to be demonstrated as table
    :param ax: Axes object where bar chart will be added
    :param rows_per_page: Rows to be shown in 1 page
    :return:
    """
    ax = ax or plt.gca()

    rows = data_frame.values.tolist()

    headers = data_frame.columns

    cell_texts = [[_wrap_text_for_table(txt) for txt in txts] for txts in rows[0:rows_per_page]]

    table = ax.table(
        cellText=cell_texts,
        colLabels=headers,
        colColours=[ORG_COLOR_HEX] * len(headers),
        loc='center',
        cellLoc='left')
    table.PAD = 0
    table.auto_set_font_size(False)
    table.set_fontsize(8)
    table.auto_set_column_width(col=list(range(len(data_frame.columns))))
    header_cell_indices = [(0, i) for i in range(len(data_frame.columns))]
    cells = table.get_celld()
    table.scale(1, 2)
    for cell in cells.values():
        cell.PAD = 0.03
    for header_cell in header_cell_indices:
        cells[header_cell].set_text_props(color="white")

    pagin_table = TablePagination(ax, table, rows, rows_per_page=rows_per_page)

    return pagin_table


def set_date_ticker(ax: Axes):
    """Adjusts x-axis ticks for time series plots"""
    ax = ax or plt.gca()
    locator = mdates.AutoDateLocator()
    formatter = mdates.ConciseDateFormatter(locator)
    formatter.formats = ['%Y',
                         '%b',
                         '%d',
                         '%H:%M',
                         '%H:%M:%S',
                         '%H:%M:%S.%f',
                         ]

    formatter.zero_formats = [''] + formatter.formats[:-1]

    formatter.zero_formats[3] = '%d-%b'

    formatter.offset_formats = ['',
                                '%Y',
                                '%b %Y',
                                '%d %b %Y',
                                '%d %b %Y',
                                '%d %b %Y']
    ax.xaxis.set_major_locator(locator)
    ax.xaxis.set_major_formatter(formatter)
    ax.fmt_xdata = mdates.DateFormatter('%Y-%m-%d %H:%M:%S')


def _wrap_text_for_table(txt: str, max_line_length=30) -> str:
    return '\n'.join(textwrap.wrap(txt, max_line_length, break_long_words=True))


def wrap_labels(ax, width, break_long_words=False, axis="x"):
    """Wraps text labels in given axis"""
    labels = []
    if axis == "x":
        for label in ax.get_xticklabels():
            text = label.get_text()
            labels.append(textwrap.fill(text, width=width,
                                        break_long_words=break_long_words))
        ax.set_xticklabels(labels, rotation=0)
    if axis == "y":
        y_labels = []
        for label in ax.get_yticklabels():
            text = label.get_text()
            y_labels.append(textwrap.fill(text, width=width,
                                          break_long_words=break_long_words))
        ax.set_yticklabels(y_labels, rotation=0)



"""
This module includes AfeTargetColumn class.
"""
from organon.afe.domain.common.reader_helper import get_values_from_kwargs
from organon.afe.domain.enums.afe_target_column_type import AfeTargetColumnType
from organon.afe.domain.settings.binary_target import BinaryTarget


class AfeTargetColumn:
    """
    Information for target column of Automated Feature Extraction process
    """
    ATTR_DICT = {
        "column_name": str,
        "target_column_type": AfeTargetColumnType,
        "binary_target_info": BinaryTarget
    }

    def __init__(self, **kwargs):
        self.column_name: str = None
        self.target_column_type: AfeTargetColumnType = None
        self.binary_target_info: BinaryTarget = None

        get_values_from_kwargs(self, self.ATTR_DICT, kwargs)



"""Includes CrossValidation class"""
from typing import List

import pandas as pd

from organon.ml.modelling.algorithms.core.abstractions.base_modeller import BaseModeller
from organon.ml.modelling.model_selection.domain.objects.cross_validation_output import CrossValidationOutput
from organon.ml.modelling.model_selection.domain.services.cross_validation_service import CrossValidationService, \
    DEFAULT_CV_COUNT
from organon.ml.modelling.model_selection.settings.model_selection_user_input_service import \
    ModelSelectionUserInputService


class CrossValidation:
    """Class for comparison of multiple modellers after fitting with the cross validation"""

    def __init__(self, modellers: List[BaseModeller], train_data: pd.DataFrame,
                 target_data: pd.Series, cv_count: int = DEFAULT_CV_COUNT, bin_count=None, return_test_fold=False):
        self.settings = ModelSelectionUserInputService.get_cross_validation_settings(modellers, train_data,
                                                                                     target_data, cv_count, bin_count,
                                                                                     return_test_fold)
        self.output: CrossValidationOutput = None

    def execute(self):
        """Fits modellers and returns scores for every run of every modeller"""
        output = CrossValidationService.execute(self.settings)
        self.output = output
        return output



"""Includes MultiTargetModellingHelper class."""
import contextlib
import random
from typing import List, Optional

import numpy as np
import pandas as pd
from lightgbm import LGBMClassifier, LGBMRegressor
from sklearn.metrics import roc_curve, auc

from organon.afe.domain.enums.afe_operator import AfeOperator
from organon.afe.domain.enums.afe_target_column_type import AfeTargetColumnType
from organon.afe.domain.modelling.businessobjects.data_frame_builder import DataFrameBuilder
from organon.afe.domain.modelling.supervised.afe_supervised_algorithm_settings import AfeSupervisedAlgorithmSettings
from organon.afe.domain.modelling.supervised.multi_target_entity_container_service import \
    MultiTargetEntityContainerService
from organon.afe.domain.settings.base_afe_modelling_settings import BaseAfeModellingSettings
from organon.afe.domain.settings.model_settings import ModelSettings
from organon.afe.domain.settings.target_descriptor import TargetDescriptor
from organon.fl.core.businessobjects.dict_dataframe import DictDataFrame
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.core.businessobjects.data_partition import DataPartition
from organon.fl.core.businessobjects.idata_partition import IDataPartition
from organon.fl.logging.helpers.log_helper import LogHelper


class MultiTargetModellingHelper:
    """Helper class for MultiTarget modelling operations."""

    def __init__(self, settings: BaseAfeModellingSettings):
        self._settings = settings

    @staticmethod
    def get_partition(frame_builder: DataFrameBuilder = None,
                      algorithm_settings: AfeSupervisedAlgorithmSettings = None) -> Optional[IDataPartition]:
        """Get partition for frame builder data"""
        partition = DataPartition(frame_builder.frame)
        partition.partition(algorithm_settings.training_percentage)
        return partition

    def get_selected_cols(self, num_threads: int,
                          partition: IDataPartition = None,
                          algorithm_settings: AfeSupervisedAlgorithmSettings = None,
                          frame_builder: DataFrameBuilder = None,
                          frame_with_all_columns: DictDataFrame = None,
                          is_final=False) -> List[str]:
        """Selects best afe columns by modelling"""
        target_descriptor = self._settings.data_source_settings.target_descriptor

        categorical_features = self.__get_categorical_features(frame_builder, frame_with_all_columns.get_column_names())

        model = MultiTargetModellingHelper.__build_model(num_threads,
                                                         frame_builder.target_array,
                                                         target_descriptor, partition,
                                                         frame_with_all_columns,
                                                         algorithm_settings, categorical_features, is_final)
        selected_cols = []
        if model is not None and not model.empty:
            selected_cols.extend(model["variable"])

        return selected_cols

    @staticmethod
    def __get_categorical_features(frame_builder: DataFrameBuilder, column_names: List[str] = None) \
            -> List[str]:
        categorical_features = []
        for afe_column_name in column_names:
            afe_column = frame_builder.name_to_column[afe_column_name]
            if AfeOperator.Mode.name == afe_column.operator.name:
                categorical_features.append(afe_column_name)
        return categorical_features

    # pylint: disable=too-many-arguments
    @staticmethod
    def __build_model(num_of_threads: int,
                      target_array: np.array,
                      target_descriptor: TargetDescriptor,
                      partition: IDataPartition,
                      frame_with_all_columns: DictDataFrame,
                      algorithm_settings: AfeSupervisedAlgorithmSettings,
                      categorical_feature_list: List[str],
                      is_final=False):
        if is_final:
            model_settings: ModelSettings = algorithm_settings.final_model_settings
        else:
            model_settings: ModelSettings = algorithm_settings.model_settings

        y_all = target_array
        train_cols = frame_with_all_columns.get_column_names()
        if len(train_cols) == 0:
            raise KnownException("There are no frame columns to build model!")
        x_all = frame_with_all_columns.get_subset_as_pandas_df(columns=train_cols)

        for feature in categorical_feature_list:
            x_all[feature] = x_all[feature].astype("category")

        lgbm = MultiTargetModellingHelper.__get_lgbm_modeller(target_descriptor.target_column.target_column_type,
                                                              model_settings, num_of_threads,
                                                              categorical_feature_list)

        if partition.validation_indices is None:
            raise KnownException(
                "Validation set cannot be empty for lightgbm classifier. Please update training_percentage")
        with contextlib.redirect_stdout(None):
            lgbm.fit(x_all.iloc[partition.training_indices], y_all[partition.training_indices],
                     eval_set=[[x_all.iloc[partition.validation_indices], y_all[partition.validation_indices]]],
                     **model_settings.model_fit_params, verbose=0)

        imp_df_lgbm = pd.DataFrame(lgbm.feature_importances_, columns=['importance'], index=x_all.columns)
        imp_df_lgbm['imp_ratio'] = imp_df_lgbm['importance'] / imp_df_lgbm['importance'].sum()
        imp_df_lgbm.sort_values(by=['imp_ratio'], ascending=[False], inplace=True)
        imp_df_lgbm['cum_importance_perc'] = imp_df_lgbm['imp_ratio'].cumsum()
        if imp_df_lgbm['cum_importance_perc'].iloc[0] >= model_settings.reduction_coverage:
            imp_df_lgbm = imp_df_lgbm.iloc[[0]]
        else:
            imp_df_lgbm = imp_df_lgbm[imp_df_lgbm['cum_importance_perc'] <= model_settings.reduction_coverage]
        imp_df_lgbm = imp_df_lgbm.reset_index()[['index', 'importance']].rename(columns={'index': 'variable'})
        return imp_df_lgbm

    def get_final_columns_metrics(self, num_threads: int, frame_builder: DataFrameBuilder,
                                  frame_with_all_columns: DictDataFrame, reduced_columns: List[str],
                                  algorithm_settings: AfeSupervisedAlgorithmSettings) -> dict:
        """Returns modelling metrics as a dictionary"""
        y_all = frame_builder.target_array

        model_settings = algorithm_settings.final_model_settings
        target_descriptor = self._settings.data_source_settings.target_descriptor

        categorical_features = self.__get_categorical_features(frame_builder, reduced_columns)

        x_all = frame_with_all_columns.get_subset_as_pandas_df(columns=reduced_columns)

        for feature in categorical_features:
            x_all[feature] = x_all[feature].astype("category")

        train_indices, validation_indices, test_indices = self.__get_indices_for_final_col_metrics_model(
            len(x_all),
            algorithm_settings.training_percentage
        )
        lgbm = self.__get_lgbm_modeller(target_descriptor.target_column.target_column_type, model_settings,
                                        num_threads, categorical_features)
        with contextlib.redirect_stdout(None):
            lgbm.fit(x_all.iloc[train_indices], y_all[train_indices],
                     eval_set=[[x_all.iloc[validation_indices], y_all[validation_indices]]],
                     **model_settings.model_fit_params, verbose=0)

        y_test = y_all[test_indices]
        fpr, tpr, auc_val = {}, {}, {}
        target_col = target_descriptor.target_column
        if target_col.target_column_type == AfeTargetColumnType.Binary:
            y_pred = lgbm.predict_proba(x_all.iloc[test_indices])[:, 1]
            fpr["all"], tpr["all"], _ = roc_curve(y_test, y_pred,
                                                  pos_label=target_col.binary_target_info.positive_category)

            auc_val["all"] = auc(fpr["all"], tpr["all"])
        elif target_descriptor.target_column.target_column_type == AfeTargetColumnType.MultiClass:
            classes = lgbm.classes_
            y_pred = lgbm.predict_proba(x_all.iloc[test_indices])
            y_test_one_hot = np.ndarray((len(test_indices), len(classes)))
            for i, cls in enumerate(classes):
                y_test_one_hot[:, i] = np.array(y_test == cls).astype(int)
                fpr[cls], tpr[cls], _ = roc_curve(y_test_one_hot[:, i], y_pred[:, i])
                auc_val[cls] = auc(fpr[cls], tpr[cls])
            fpr["micro"], tpr["micro"], _ = roc_curve(y_test_one_hot.ravel(), y_pred.ravel())
            auc_val["micro"] = auc(fpr["micro"], tpr["micro"])
            all_fpr = np.unique(np.concatenate([fpr[cls] for cls in classes]))
            mean_tpr = np.zeros_like(all_fpr)
            for cls in classes:
                mean_tpr += np.interp(all_fpr, fpr[cls], tpr[cls])
            mean_tpr /= len(classes)
            fpr["macro"] = all_fpr
            tpr["macro"] = mean_tpr
            auc_val["macro"] = auc(fpr["macro"], tpr["macro"])

        feature_importance = pd.DataFrame(sorted(zip(lgbm.feature_importances_, x_all.columns), reverse=True),
                                          columns=['Value', 'Feature'])

        return {"fpr": fpr, "tpr": tpr, "auc": auc_val, "feature_importances": feature_importance}

    @staticmethod
    def __get_indices_for_final_col_metrics_model(frame_length: int, training_percentage: float):
        system_random = random.SystemRandom()
        if training_percentage >= 1.0:
            training_ratio = 0.6
            validation_ratio = 0.2
        else:
            training_ratio = training_percentage ** 2
            validation_ratio = training_percentage - training_ratio
        t_list = []
        v_list = []
        test_list = []
        for i in range(frame_length):
            rand_num = system_random.uniform(0, 1)
            if rand_num <= training_ratio:
                t_list.append(i)
            elif rand_num <= training_ratio + validation_ratio:
                v_list.append(i)
            else:
                test_list.append(i)
        training_indices = t_list
        validation_indices = v_list
        test_indices = test_list

        return training_indices, validation_indices, test_indices

    def control_target_record_files(self, multi_target_entity_container_service: MultiTargetEntityContainerService):
        """Controls min_data_in_leaf an sample_size values and logs a warning if necessary"""
        record_counts = []
        for record in multi_target_entity_container_service.container.records_per_file.values():
            record_counts.append(record.actual_record_count)
        min_sample_size = min(record_counts)

        final_model_settings: ModelSettings = self._settings.algorithm_settings.final_model_settings

        model_settings: ModelSettings = self._settings.algorithm_settings.model_settings
        min_data_in_leaf_and_sample_size_control_ratio = \
            self._settings.algorithm_settings.min_data_in_leaf_and_sample_size_control_ratio

        if (final_model_settings.model_params["min_data_in_leaf"] >
                min_sample_size * min_data_in_leaf_and_sample_size_control_ratio or
                model_settings.model_params["min_data_in_leaf"] >
                min_sample_size * min_data_in_leaf_and_sample_size_control_ratio):
            LogHelper.warning("Entered min_data_in_leaf is too large for the sample size "
                              "so model will not be established.")

    @staticmethod
    def __get_lgbm_modeller(target_column_type: AfeTargetColumnType, model_settings: ModelSettings,
                            num_of_threads: int, categorical_feature_list: List[str]):
        n_jobs = num_of_threads
        if n_jobs is None or n_jobs == 0:
            n_jobs = -1
        if target_column_type == AfeTargetColumnType.Binary:
            model_settings.model_fit_params.setdefault('eval_metric', 'auc')
            model_settings.model_params.setdefault('objective', 'binary')
            lgbm = LGBMClassifier(n_jobs=n_jobs, categorical_feature=categorical_feature_list,
                                  **model_settings.model_params)
        elif target_column_type == AfeTargetColumnType.Scalar:
            model_settings.model_fit_params.setdefault('eval_metric', 'mape')
            lgbm = LGBMRegressor(n_jobs=n_jobs, categorical_feature=categorical_feature_list,
                                 **model_settings.model_params)
        elif target_column_type == AfeTargetColumnType.MultiClass:
            model_settings.model_fit_params.setdefault('eval_metric', 'multi_logloss')
            model_settings.model_params.setdefault('objective', 'multiclass')
            lgbm = LGBMClassifier(n_jobs=n_jobs, categorical_feature=categorical_feature_list,
                                  **model_settings.model_params)
        else:
            raise KnownException("Unknown target type :" + target_column_type)
        return lgbm



"""
This module keeps FlCoreStaticObjects class.
"""
from typing import List

from organon.fl.core.fileoperations import directory_helper
from organon.fl.core.helpers import string_helper
from organon.fl.core.iocutil.ioc_registration_item import IocRegistrationItem


class FlCoreStaticObjects:
    """
    Class for holding ioc registered items, registered connections and application_initialize_called boolean object.
    """
    ioc_registered_items: List[IocRegistrationItem] = None
    application_initialize_called: bool = None

    fl_app_output_directory: str = None

    @staticmethod
    def get_master_output_directory():
        """Returns master output directory for application"""
        FlCoreStaticObjects.set_fl_app_context_from_config()
        if not string_helper.is_null_or_empty(FlCoreStaticObjects.fl_app_output_directory):
            return FlCoreStaticObjects.fl_app_output_directory
        return directory_helper.get_current_directory()

    @staticmethod
    def set_fl_app_context_from_config():
        """Sets fl context settings from config file."""



"""Includes abstract class(IDataFrame) for custom DataFrame classes."""
import abc
from typing import List, Any, Union


class IDataFrame(metaclass=abc.ABCMeta):
    """Abstract base class for custom DataFrame classes."""

    def __init__(self, row_count: int = None):
        self.data_frame = None
        self.row_count = row_count

    @abc.abstractmethod
    def set_value(self, value, column_name: str, row_index: int):
        """Sets value of a cell in the given column and row"""
        raise NotImplementedError

    @abc.abstractmethod
    def contains_column(self, column_name: str) -> bool:
        """Checks if column is in dataframe"""
        raise NotImplementedError

    @abc.abstractmethod
    def try_add(self, column_name: str, data: List[Any], mark_for_garbage_collection: bool = False):
        """Sets given data as value of given column"""
        raise NotImplementedError

    @abc.abstractmethod
    def remove(self, column_names: Union[List[str], str]):
        """Removes given columns from dataframe"""
        raise NotImplementedError

    @abc.abstractmethod
    def collect_garbage_columns(self):
        """Removes columns marked for garbage collection from dataframe"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_value(self, column_name: str, row_index: int = None):
        """Returns all values in column if row_index is None. Else, return value at given column and row."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_column_names(self) -> List[str]:
        """Returns all column names"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_subset_as_pandas_df(self, indices: List[Any] = None, columns: List[str] = None):
        """Returns a subset of dataframe which only includes given columns and given indices."""
        raise NotImplementedError

    @abc.abstractmethod
    def copy_as_pandas_df(self):
        """Returns a copy of dataframe as pandas dataframe"""
        raise NotImplementedError



"""Module for LinearRegressionAlgorithmSettings"""
from dataclasses import dataclass
import numpy as np

from organon.fl.mathematics.sweep.enums import RegressionAttributeSelectionMethod


@dataclass
class LinearRegressionAlgorithmSettings:
    """Settings for the Linear Regression"""
    target: int
    weight: np.ndarray
    attribute_selection_method: RegressionAttributeSelectionMethod
    inclusion_confidence_level: float
    exclusion_confidence_level: float
    inclusion_f_statistics: float
    exclusion_f_statistics: float
    use_positive_coefficients: bool
    use_intercept: bool
    min_r_squared_change_in_train_set: float
    min_r_squared_change_in_validation_set: float
    maximum_vif: float
    tolerance: float



"""Module for SelectedAttribute"""
from dataclasses import dataclass

from organon.fl.mathematics.sweep.enums import AttributeSelectionStatus, TestStatisticsType


@dataclass
class SelectedAttribute:
    """Selected Attribute info"""
    attribute_name: str
    selection_status: AttributeSelectionStatus
    selection_test_statistics: TestStatisticsType
    selection_statistics: float
    updated_training_performance: float
    updated_validation_performance: float
    performance_change_in_training_set: float
    performance_change_in_validation_set: float

    def __str__(self) -> str:
        result_string = f"Attribute Name: {self.attribute_name}\n"
        result_string += f"Selection Status: {self.selection_status}\n"
        result_string += f"Selection Test Statistics: {self.selection_test_statistics}\n"
        result_string += f"Selection Statistics: {self.selection_statistics}\n"
        result_string += f"Updated Training Performance: {self.updated_training_performance}\n"
        result_string += f"Updated Validation Performance: {self.updated_validation_performance}\n"
        result_string += f"Performance Change In Training-Set: {self.performance_change_in_training_set}\n"
        result_string += f"Performance Change In Validation-Set: {self.performance_change_in_validation_set}\n"

        return result_string



"""Includes RecordSourceFilterHelper class."""
from collections import abc
from typing import List

import numpy as np
import pandas as pd
from pandas.io.parsers import TextFileReader

from organon.fl.core.enums.datetime_token import DatetimeToken
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.core.helpers import list_helper
from organon.idq.domain.settings.abstractions.dq_base_input_source_settings import DqBaseInputSourceSettings
from organon.idq.domain.settings.date_value_definition import DateValueDefinition
from organon.idq.domain.settings.input_source.dq_file_input_source_settings import DqFileInputSourceSettings
from organon.idq.domain.settings.partition_info import PartitionInfo


class FilteredTextFileReader(abc.Iterator):
    """Iterator class for iterating csv file using TextFileReader and returning rows after filtering"""

    def __init__(self, text_file_reader: TextFileReader, settings: DqFileInputSourceSettings):
        self._text_file_reader = text_file_reader
        self._settings = settings

    def __next__(self):
        data_frame = self._text_file_reader.__next__()
        return RecordSourceFilterHelper.get_filtered_dataframe(data_frame, self._settings)


class RecordSourceFilterHelper:
    """Includes helper functions for reading a record source with filter"""

    @classmethod
    def read_csv_with_chunks(cls, settings: DqFileInputSourceSettings) -> FilteredTextFileReader:
        """Returns pandas TextFileReader object for reading with chunks according to given DqFileInputSourceSettings"""
        record_source: str = settings.source.locator
        chunk_size = settings.number_of_rows_per_step
        included_columns = settings.included_columns
        sep = settings.csv_separator
        date_columns = settings.date_columns
        if included_columns is None:
            dfs = pd.read_csv(record_source, sep=sep, chunksize=chunk_size, parse_dates=date_columns)
        else:
            dfs = pd.read_csv(record_source, sep=sep, chunksize=chunk_size, usecols=included_columns,
                              parse_dates=date_columns)
        return FilteredTextFileReader(dfs, settings)

    @classmethod
    def get_filtered_dataframe(cls, data_frame: pd.DataFrame, settings: DqBaseInputSourceSettings) -> pd.DataFrame:
        """Filters given dataframe according to given input source settings"""
        data_frame = cls._filter_by_partition(data_frame, settings.partition_info_list)
        if settings.filter_callable is not None:
            try:
                data_frame = settings.filter_callable(data_frame)
            except Exception as exc:
                raise KnownException("Error occurred while executing 'filter_callable'") from exc
            if not isinstance(data_frame, pd.DataFrame):
                raise KnownException(f"'filter_callable' is expected to return a pandas DataFrame. "
                                     f"Returned: {type(data_frame).__name__}")
        return data_frame

    @classmethod
    def _filter_by_partition(cls, data_frame: pd.DataFrame, partition_info_list: List[PartitionInfo]) -> pd.DataFrame:
        """Filters given dataframe according to given partition info"""
        if list_helper.is_null_or_empty(partition_info_list):
            return data_frame
        all_filter = cls.__get_filter(data_frame, partition_info_list)
        return data_frame[all_filter]

    @classmethod
    def __get_filter(cls, data_frame: pd.DataFrame, partition_info_list: List[PartitionInfo]) -> np.array:
        """Returns df filter as pd.Series according to given partition info"""
        all_filter = np.array([True for _ in range(len(data_frame))])
        for partition_info in partition_info_list:
            is_null_filter = None
            if None in partition_info.column_values:
                is_null_filter = data_frame[partition_info.column_name].isnull().to_numpy()

            non_null_column_values = [value for value in partition_info.column_values if value is not None]

            _filter = cls.__get_filter_for_partition(data_frame, partition_info.column_name,
                                                     non_null_column_values)
            if is_null_filter is not None:
                _filter |= is_null_filter
            all_filter &= _filter
        return all_filter

    @classmethod
    def __get_filter_for_partition(cls, data_frame: pd.DataFrame, column_name: str,
                                   non_null_column_values: list) -> np.array:
        if len(non_null_column_values) == 0:
            return np.array([True for _ in range(len(data_frame))])
        first_val = non_null_column_values[0]
        if isinstance(first_val, DateValueDefinition):
            format_val_dict = {}
            for val in non_null_column_values:
                conversion_format = cls.__get_str_conversion_format(val.get_datetime_tokens())
                val_str = cls._get_column_value_string_for_date_column(val)
                if conversion_format in format_val_dict:
                    format_val_dict[conversion_format].append(val_str)
                else:
                    format_val_dict[conversion_format] = [val_str]
            date_filter = np.array([False for _ in range(len(data_frame))])
            for conversion_format, value_strings in format_val_dict.items():
                date_filter |= data_frame[column_name].dt.strftime(conversion_format).isin(value_strings).to_numpy()
            return date_filter
        return data_frame[column_name].isin(non_null_column_values).to_numpy()

    @classmethod
    def _get_column_value_string_for_date_column(cls, column_value: DateValueDefinition):
        date_str = ""
        if column_value.year is not None:
            date_str += ("000" + str(column_value.year))[-4:]
        if column_value.month is not None:
            date_str += ("0" + str(column_value.month))[-2:]
        if column_value.day is not None:
            date_str += ("0" + str(column_value.day))[-2:]
        if column_value.hour is not None:
            date_str += ("0" + str(column_value.hour))[-2:]
        return date_str

    @classmethod
    def __get_str_conversion_format(cls, datetime_tokens: List[DatetimeToken]) -> str:
        conversion_str = ""
        if DatetimeToken.YEAR in datetime_tokens:
            conversion_str += "%Y"
        if DatetimeToken.MONTH in datetime_tokens:
            conversion_str += "%m"
        if DatetimeToken.DAY in datetime_tokens:
            conversion_str += "%d"
        if DatetimeToken.HOUR in datetime_tokens:
            conversion_str += "%H"
        return conversion_str
